# Copyright (C) 2006-2014 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import time

from agent.api.base import Agent
from agent.lib.constants import (LOCAL_SERVER_DB,
                                 MEMBERSHIP_STATE_HB_CONFIGURED,
                                 MEMBERSHIP_STATE_CLUSTERED)
from agent.lib.db import (read_item, write_item, delete_item, dump_db,
                          truncate_db, create_db, update_db,
                          get_cluster_db_home)
from agent.lib.notification import notify_manager
from agent.lib.settings import get_membership_state, get_poolfs_uuid
from agent.target.cluster import is_cluster_online


class Cluster(Agent):
    """
    Class for cluster management.
    """

    def configure_server_for_cluster(self, o2cb_conf, cluster_conf,
                poolfs_type, poolfs_target, poolfs_uuid, poolfs_nfsbase_uuid):
        """
        Configures the server for the cluster.

        @param o2cb_conf: system parameters for the cluster
        @type  o2cb_conf: C{dict}

        @param cluster_conf: the cluster configuration
        @type  cluster_conf: C{str}

        @param poolfs_type: type of the pool filesystem, either C{'nfs'}
        or C{'lun'}
        @type  poolfs_type: C{str}

        @param poolfs_target: physical location of the pool filesystem
        @type  poolfs_target: C{str}

        @param poolfs_uuid: a unique identifier for the pool filesystem
        @type  poolfs_uuid: C{str}

        @param poolfs_nfsbase_uuid: a unique identifier for the base NFS
        filesystem
        @type  poolfs_nfsbase_uuid: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()

    def deconfigure_server_for_cluster(self, poolfs_uuid):
        """
        Removes the cluster configuration.

        @param poolfs_uuid: a unique identifier for the pool filesystem
        @type  poolfs_uuid: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()

    def update_cluster_configuration(self, cluster_conf):
        """
        Updates the cluster configuration.

        @param cluster_conf: the cluster configuration
        @type  cluster_conf: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()

    def create_cluster(self, poolfs_uuid):
        """
        Creates a new cluster and add this server to the cluster.

        @param poolfs_uuid: pool filesystem uuid
        @type  poolfs_uuid: C{str}

        @raise IModuleException: on failure

        """
        validate_cluster_setup(poolfs_uuid, True)
        cluster_add_server(poolfs_uuid, True)

    def join_cluster(self, poolfs_uuid):
        """
        Add this server to an existing cluster.

        @param poolfs_uuid: pool filesystem uuid
        @type  poolfs_uuid: C{str}

        @raise IModuleException: on failure

        """
        validate_cluster_setup(poolfs_uuid, False)
        cluster_add_server(poolfs_uuid, False)

    def destroy_cluster(self, poolfs_uuid, force=False):
        """
        Destroy the cluster.

        @param poolfs_uuid: pool filesystem uuid
        @type  poolfs_uuid: C{str}

        @raise IModuleException: on failure

        """
        validate_cluster_teardown(poolfs_uuid, True)
        try:
            cluster_remove_server(poolfs_uuid, True)
        except:
            msg = 'Failed to destroy cluster'
            logging.exception(msg)
            send_notification('{CLUSTER} {WARNING} %s' % msg)
            write_item(LOCAL_SERVER_DB, 'clustered', False)

    def leave_cluster(self, poolfs_uuid, force=False):
        """
        Removes this server from the cluster.

        @param poolfs_uuid: pool filesystem uuid
        @type  poolfs_uuid: C{str}

        @raise IModuleException: on failure

        """
        validate_cluster_teardown(poolfs_uuid, False)
        try:
            cluster_remove_server(poolfs_uuid, False)
        except:
            msg = 'Failed to leave cluster'
            logging.exception(msg)
            send_notification('{CLUSTER} {WARNING} %s' % msg)
            write_item(LOCAL_SERVER_DB, 'clustered', False)

    def is_cluster_online(self):
        """
        Indicate if the cluster is online or not.

        @return: C{True} if the cluster is online, C{False} otherwise.
        @rtype: C{bool}

        """
        return is_cluster_online()

    def discover_cluster(self):
        """
        Discover information about the cluster configured for this server.

        @return: Returns the discovery data as an XML document
        @rtype: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


def validate_cluster_setup(poolfs_uuid, init=False):
    # check for membership state
    state = get_membership_state()
    if state == MEMBERSHIP_STATE_CLUSTERED:
        return
    elif state != MEMBERSHIP_STATE_HB_CONFIGURED:
        raise Exception("Cannot add server to cluster "
                        "in membership state %s" % state)
    # get db dict
    d = dump_db(LOCAL_SERVER_DB)

    # check for poolfs uuid
    if poolfs_uuid != d.get("poolfs_uuid"):
        raise Exception("Pool filesystem UUID does not match "
                                            "the db record.")
    # validate server roles
    master_flag = d.get("is_master")
    if init and not master_flag:
        raise Exception("Initial server must be server pool master.")
    if not init and master_flag:
        raise Exception("Server pool master must be initial server.")


def cluster_add_server(poolfs_uuid, init=False):
    """
    Adds a new server to the cluster.

    @param      poolfs_uuid - pool filesystem uuid
    @param      init - True if it is the first cluster member

    @return     None. Raises an exception on failure.
    """
    # get db dict
    d = dump_db(LOCAL_SERVER_DB)

    # hostname and server info
    hostname = d.get("registered_hostname")
    srv = {
        "roles": d.get("roles"),
        "node_number": d.get("node_number"),
        "registered_ip": d.get("registered_ip"),
        "is_master": d.get("is_master")}

    db_home = get_cluster_db_home()
    if init:
        # shared db initialization
        create_db("server_pool_servers", True, db_home)
        write_item("server_pool_servers", hostname, srv, db_home)
        create_db("server_pool", True, db_home)
        pool_info = {
            "pool_alias": d.get("pool_alias"),
            "pool_uuid": d.get("pool_uuid"),
            "pool_virtual_ip": d.get("pool_virtual_ip"),
            "pool_master_hostname": hostname,
            "pool_member_ip_list": d.get("pool_member_ip_list"),
            "auto_remaster": True}
        update_db("server_pool", pool_info, db_home)
        create_db("monitored_vms", True, db_home)
    else:
        # update server information
        write_item("server_pool_servers", hostname, srv, db_home)

    # remove pool_member_ip_list from local db
    if "pool_member_ip_list" in d:
        delete_item(LOCAL_SERVER_DB, "pool_member_ip_list")

    # mark server as clustered
    write_item(LOCAL_SERVER_DB, "clustered", True)


def validate_cluster_teardown(poolfs_uuid, final=False):
    # check for membership state
    state = get_membership_state()
    if state == MEMBERSHIP_STATE_HB_CONFIGURED:
        return
    elif state != MEMBERSHIP_STATE_CLUSTERED:
        raise Exception("Cannot remove server from cluster "
                        "in membership state %s" % state)
    # check for poolfs uuid
    uuid = get_poolfs_uuid()
    if poolfs_uuid != uuid:
        raise Exception("Pool filesystem UUID does not match "
                                            "the db record.")
    # get db dict
    d = dump_db(LOCAL_SERVER_DB)

    # validate server roles
    master_flag = d.get("is_master")
    if final and not master_flag:
        raise Exception("Final server in the cluster must be the pool master.")
    if not final and master_flag:
        raise Exception("Pool master cannot be removed from the cluster.")


def cluster_remove_server(poolfs_uuid, final=False):
    """
    Removes a server from the cluster.

    @param      poolfs_uuid - pool filesystem uuid
    @param      final - True if it is the final server

    @return     None. Raises an exception on failure.
    """
    # get db dict
    d = dump_db(LOCAL_SERVER_DB)

    # update local db
    db_home = get_cluster_db_home()
    ip_list = read_item("server_pool", "pool_member_ip_list", db_home)
    update = {
        "clustered": False,
        "pool_member_ip_list": ip_list}
    update_db(LOCAL_SERVER_DB, update)

    try:
        if final:
            # wait before truncating db
            time.sleep(5)
            # truncate db
            truncate_db("server_pool", db_home)
            truncate_db("server_pool_servers", db_home)
        else:
            # remove db record
            delete_item("server_pool_servers", d.get("registered_hostname"),
                        db_home)
    except Exception, e:
        try:
            # restore local db
            update = {
                "clustered": True,
                "pool_member_ip_list": None}
            update_db(LOCAL_SERVER_DB, update)
        finally:
            raise e


def send_notification(msg):
    try:
        notify_manager(msg)
    except:
        logging.exception('Failed to send notification')
