# Copyright (C) 2006-2014 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import string
import logging
from agent.api.cluster.base import Cluster
from agent.lib import filelock
from agent.lib.constants import (LOCAL_SERVER_DB,
                                 MEMBERSHIP_STATE_HB_CONFIGURED,
                                 MEMBERSHIP_STATE_CLUSTERED,
                                 MEMBERSHIP_STATE_POOLED,
                                 NFS_MOUNT,
                                 POOLFS_MOUNT, POOLFS_META)
from agent.lib.db import read_item, dump_db, update_db
from agent.lib.discover import SimpleDiscoveryXML
from agent.lib.files import rename_file, file_read
from agent.lib.settings import (get_membership_state, get_poolfs_uuid,
                                parse_metafile)
from agent.target.cluster import (CLUSTER_CONF, CLUSTER_CONF_DIR,
                                  CLUSTER_CONF_ORIG, CLUSTER_CONFIGFS,
                                  O2CB_CONF, POOLFS_IMG,
                                  mount_device, umount_device,
                                  setup_loop_device, detach_loop_device,
                                  get_ocfs2_cluster_name, ocfs2_file_read,
                                  server_cluster_online, server_cluster_offline)
from agent.target.host import run_cmd
from agent.lib.ocfs2 import get_cluster_interface, find_associated_loop_devices


class ClusterO2CB(Cluster):
    """
    Class for cluster management.
    """

    def get_services(self, version=None):
        return {
            'configure_server_for_cluster':
            self.configure_server_for_cluster,
            'deconfigure_server_for_cluster':
            self.deconfigure_server_for_cluster,
            'update_cluster_configuration':
            self.update_cluster_configuration,
            'create_cluster': self.create_cluster,
            'join_cluster': self.join_cluster,
            'destroy_cluster': self.destroy_cluster,
            'leave_cluster': self.leave_cluster,
            'is_cluster_online': self.is_cluster_online,
            'discover_cluster': self.discover_cluster,
        }

    def configure_server_for_cluster(self, o2cb_conf, cluster_conf,
                poolfs_type, poolfs_target, poolfs_uuid, poolfs_nfsbase_uuid):
        """
        Configures the server for the cluster.

        @param      o2cb_conf - dict of o2cb sysconfig
        @param      cluster_conf - cluster configuration file
        @param      poolfs_type - "nfs" or "lun"
        @param      poolfs_target -
                    physical location of the pool filesystem
        @param      poolfs_uuid -
                    a unique identifier for the pool filesystem
        @param      poolfs_nfsbase_uuid -
                    a unique identifier for the base NFS filesystem

        @return     None. Raises an exception on failure.
        """
        if poolfs_type not in ("nfs", "lun"):
            raise ValueError("Invalid pool filesystem type: %s" % poolfs_type)

        # check for membership state
        state = get_membership_state()
        if state != MEMBERSHIP_STATE_POOLED:
            raise Exception("Cannot configure server for cluster in membership state %s" % state)

        # check for cluster.conf and cluster layout in configfs
        if os.path.exists(CLUSTER_CONF):
            raise Exception("/etc/ocfs2/cluster.conf already exists.")

        cluster_name = get_ocfs2_cluster_name()
        if cluster_name:
            raise Exception("Already in cluster: %s" % cluster_name)

        # check for poolfs uuid
        if poolfs_type == "nfs":
            # mount poolfs nfsbase
            mount_point = os.path.join(NFS_MOUNT, poolfs_nfsbase_uuid)
            mount_device(poolfs_target, mount_point)
            # read metadata
            poolfs_meta = os.path.join(mount_point, POOLFS_META)
            contents = file_read(poolfs_meta)
        else:
            contents = ocfs2_file_read(poolfs_target, POOLFS_META)

        d = parse_metafile(contents)
        ovs_poolfs_uuid = d.get("OVS_POOLFS_UUID")
        if not ovs_poolfs_uuid:
            raise Exception("No pool filesystem UUID in the metafile.")
        if ovs_poolfs_uuid != poolfs_uuid:
            raise Exception("Pool filesystem UUID does not match the metafile record.")

        if not os.path.exists(CLUSTER_CONF_DIR):
            os.makedirs(CLUSTER_CONF_DIR)

        if poolfs_type == "nfs":
            poolfs_img = os.path.join(mount_point, POOLFS_IMG)
            try:
                poolfs_device = setup_loop_device(poolfs_img)
            except Exception:
                try:
                    umount_device(poolfs_target)
                except Exception:
                    pass
                raise
        else:
            poolfs_device = poolfs_target

        stage = 0
        try:
            # update o2cb config
            new_conf = {
                "O2CB_ENABLED": "true",
                "O2CB_STACK": "o2cb",
                "O2CB_BOOTCLUSTER": "ocfs2",
                "O2CB_HEARTBEAT_THRESHOLD": "31",
                "O2CB_IDLE_TIMEOUT_MS": "60000",
                "O2CB_KEEPALIVE_DELAY_MS": "2000",
                "O2CB_RECONNECT_DELAY_MS": "2000",
                }
            old_conf = get_o2cb_conf()
            new_conf.update(old_conf)
            new_conf.update(o2cb_conf)
            set_o2cb_conf(new_conf)
            stage = 1

            # write cluster.conf
            if isinstance(cluster_conf, dict):
                config_cluster_layout(cluster_conf)
            else:
                file_write(CLUSTER_CONF, cluster_conf)
            stage = 2

            # cluster online
            server_cluster_online()

            # mount poolfs
            poolfs_mount_point = os.path.join(POOLFS_MOUNT, poolfs_uuid)
            mount_device(poolfs_device, poolfs_mount_point)

            # update db
            poolfs_info = {
                "poolfs_type": poolfs_type,
                "poolfs_target": poolfs_target,
                "poolfs_uuid": poolfs_uuid,
                "poolfs_nfsbase_uuid": poolfs_nfsbase_uuid}
            update_db(LOCAL_SERVER_DB, poolfs_info)
        except Exception:
            # remove cluster.conf
            if stage > 1 and os.path.exists(CLUSTER_CONF):
                try:
                    os.remove(CLUSTER_CONF)
                except Exception:
                    pass

            # restore o2cb config
            if stage > 0:
                try:
                    set_o2cb_conf(old_conf)
                except Exception:
                    pass

            try:
                umount_device(poolfs_device)
            except Exception:
                pass

            # cluster offline
            try:
                server_cluster_offline()
            except Exception:
                pass

            if poolfs_type == "nfs":
                try:
                    detach_loop_device(poolfs_img)
                except Exception:
                    pass
                try:
                    umount_device(poolfs_target)
                except Exception:
                    pass
            raise

    def deconfigure_server_for_cluster(self, poolfs_uuid):
        """
        Removes the cluster configuration.

        @param      poolfs_uuid -
                    a unique identifier for the pool filesystem

        @return     None. Raises an exception on failure.
        """
        # check for membership state
        state = get_membership_state()
        if state != MEMBERSHIP_STATE_HB_CONFIGURED:
            raise Exception("Cannot deconfigure server for cluster in membership state %s" % state)
    
        # check for poolfs uuid
        uuid = get_poolfs_uuid()
        if not uuid:
            raise Exception("There is no db record of pool filesystem UUID.")
        if poolfs_uuid != uuid:
            raise Exception("Pool filesystem UUID does not match the db record.")
    
        # check for cluster.conf
        if not os.path.exists(CLUSTER_CONF):
            raise Exception("/etc/ocfs2/cluster.conf does not exist.")
    
        d = dump_db(LOCAL_SERVER_DB)
        poolfs_type = d.get("poolfs_type")
        poolfs_target = d.get("poolfs_target")
        poolfs_nfsbase_uuid = d.get("poolfs_nfsbase_uuid")
        poolfs_uuid = d.get("poolfs_uuid")
        poolfs_mount_point = os.path.join(POOLFS_MOUNT, poolfs_uuid)
    
        if poolfs_type == "nfs":
            mount_point = os.path.join(NFS_MOUNT, poolfs_nfsbase_uuid)
            poolfs_img = os.path.join(mount_point, POOLFS_IMG)
            associated_loop_devices = find_associated_loop_devices(poolfs_img)
            if not associated_loop_devices:
                logging.warn("No associated loop device found for poolfs image - trying to clean up cluster...")

                try:
                    cleanup_cluster_conf()
                except Exception, e:
                    logging.error("cleanup_cluster_conf failed, error %s", e)

                return
            poolfs_device = associated_loop_devices[0]
        else:
            poolfs_device = poolfs_target
    
        umount_device(poolfs_device)
    
        try:
            server_cluster_offline()
        except Exception:
            try:
                mount_device(poolfs_device, poolfs_mount_point)
            except Exception:
                pass
            raise
    
        if poolfs_type == "nfs":
            try:
                detach_loop_device(poolfs_img)
                umount_device(poolfs_target)
            except Exception:
                try:
                    setup_loop_device(poolfs_img)
                except Exception:
                    pass
                try:
                    server_cluster_online()
                except Exception:
                    pass
                try:
                    mount_device(poolfs_device, poolfs_mount_point)
                except Exception:
                    pass
                raise
    
        # backup cluster.conf
        rename_file(CLUSTER_CONF, CLUSTER_CONF_ORIG)
    
        # delete db records
        poolfs_info = {
            "poolfs_type": None,
            "poolfs_target": None,
            "poolfs_uuid": None,
            "poolfs_nfsbase_uuid": None,
        }
        update_db(LOCAL_SERVER_DB, poolfs_info)

    def update_cluster_configuration(self, cluster_conf):
        """
        Updates /etc/ocfs2/cluster.conf with cluster_conf.

        @param      cluster_conf - cluster configuration file

        @return     None. Raises an exception on failure.
        """
        state = get_membership_state()
        if state not in (MEMBERSHIP_STATE_HB_CONFIGURED, MEMBERSHIP_STATE_CLUSTERED):
            raise Exception("Cannot update cluster configuration in membership state %s" % state)

        # backup cluster.conf
        rename_file(CLUSTER_CONF, CLUSTER_CONF_ORIG)
        try:
            # update cluster.conf
            if isinstance(cluster_conf, dict):
                config_cluster_layout(cluster_conf)
            else:
                file_write(CLUSTER_CONF, cluster_conf)
        except Exception:
            try:
                # restore cluster.conf
                rename_file(CLUSTER_CONF_ORIG, CLUSTER_CONF)
            except Exception:
                pass
            raise

        server_cluster_online()


    def discover_cluster(self):
        """
        Discover information about the cluster configured for this server.

        Returns both the stored information from /etc/ocfs2/cluster.conf and
        the server's info if it is actively participating in the cluster.
        Returns an empty document if the server is not configured in a cluster.

        @return     Returns the discovery data as an XML document
                    Raises an exception on failure.
        """
        lock = filelock.FileLock('/var/run/ovs-agent/discover_cluster.lock')
        lock.acquire(wait=120)
        try:
            doc = SimpleDiscoveryXML("Discover_Cluster_Result")
            discover_o2cb_config(doc)
            if os.path.isfile(CLUSTER_CONF):
                doc.push("Cluster_Information")
                try:
                    discover_stored_cluster_info(doc)
                except Exception, e:
                    err = "Unable to discover stored cluster info: %s" % e
                    logging.exception(err)
                    raise Exception(err)
                hostname = read_item(LOCAL_SERVER_DB, "registered_hostname")
                cluster_name = get_ocfs2_cluster_name()
                if hostname and cluster_name:
                    try:
                        (num, ip_addr, ip_port) = \
                                get_active_cluster_info(cluster_name, hostname)
                    except Exception, e:
                        err = "Unable to discover active cluster info: %s" % e
                        logging.debug(err)
                        raise Exception(err)
                    if num != None:
                        doc.push("Active")
                        doc.add("Local_Node_Number", num)
                        doc.add("Local_IP_Address", ip_addr)
                        doc.add("Local_IP_Port", ip_port)
                        doc.add('Network_Interface', get_cluster_interface())
                        doc.pop() # Active
                doc.pop() # Cluster_Information
            return doc.finish()
        finally:
            lock.release()


class O2CBCluster:
    def __init__(self, cfg_file):
        self.config_file = cfg_file

    def add_cluster(self, cluster):
        """Adds cluster to the config file."""
        run_cmd(['/sbin/o2cb',
                 '--config-file=%s' % self.config_file,
                 'add-cluster',
                 cluster['name'],])

    def remove_cluster(self, cluster):
        """Removes cluster from the config file."""
        run_cmd(['/sbin/o2cb',
                 '--config-file=%s' % self.config_file,
                 'remove-cluster',
                 cluster['name'],])

    def add_node(self, node):
        """Adds a node to the cluster in the config file."""
        run_cmd(['/sbin/o2cb',
                 '--config-file=%s' % self.config_file,
                 'add-node',
                 '--ip=%s' % node['ip_address'],
                 '--port=%s' % node['ip_port'],
                 '--number=%s' % node['number'],
                 node['cluster'],
                 node['name'],])

    def remove_node(self, node):
        """Removes a node from the cluster in the config file."""
        run_cmd(['/sbin/o2cb',
                 '--config-file=%s' % self.config_file,
                 'remove-node',
                 node['cluster'],
                 node['name'],])

    def add_heartbeat(self, heartbeat):
        """Adds a heartbeat region to the cluster in the config file."""
        run_cmd(['/sbin/o2cb',
                 '--config-file=%s' % self.config_file,
                 'add-heartbeat',
                 heartbeat['cluster'],
                 heartbeat['region'],])

    def remove_heartbeat(self, heartbeat):
        """Removes a heartbeat region from the cluster in the config file."""
        run_cmd(['/sbin/o2cb',
                 '--config-file=%s' % self.config_file,
                 'remove-heartbeat',
                 heartbeat['cluster'],
                 heartbeat['region'],])

    def heartbeat_mode(self, cluster):
        """Toggles the heartbeat mode between global and local."""
        run_cmd(['/sbin/o2cb',
                 '--config-file=%s' % self.config_file,
                 'heartbeat-mode',
                 cluster['name'],
                 cluster['heartbeat_mode'],])

    def list_cluster(self):
        """Lists all the nodes and heartbeat regions associated with the
        cluster in the config file."""
        pass

    def list_nodes(self):
        """Lists all the nodes associated with the cluster in the config
        file."""
        pass

    def list_heartbeats(self):
        """Lists all the heartbeat regions associated with the cluster in the
        config file."""
        pass

    def register_cluster(self):
        """Registers the cluster with configfs."""
        pass

    def unregister_cluster(self):
        """Unregisters the cluster from configfs."""
        pass

    def start_heartbeat(self):
        """Starts global heartbeat."""
        pass

    def stop_heartbeat(self):
        """Stops global heartbeat."""
        pass

    def cluster_status(self):
        """Returns 0 if cluster online, 1 otherwise."""
        pass




def get_o2cb_conf():
    d = {}
    if os.path.exists(O2CB_CONF):
        for ln in file_readlines(O2CB_CONF):
            l = ln.split('#')[0].strip()
            if l:
                parts = map(string.strip, l.split('='))
                if len(parts) == 2 and parts[0]:
                    d[parts[0]] = parts[1]
    return d


def set_o2cb_conf(conf_dict):
    l = []
    for k, v in conf_dict.items():
        ln = '%s=%s\n' % (k, v)
        l.append(ln)
    file_writelines(O2CB_CONF, l)


def file_write(path, data):
    f = open(path, "w")
    try:
        f.write(data)
    finally:
        f.close()


def file_readlines(path):
    f = open(path, "r")
    try:
        return f.readlines()
    finally:
        f.close()

def file_writelines(path, lines):
    f = open(path, "w")
    try:
        f.writelines(lines)
    finally:
        f.close()


def parse_cluster_layout(cfg_file=CLUSTER_CONF):
    """
    @return:    {'node':[<dict of node>],
                 'cluster': [<dict of cluster>],
                 'heartbeat': [<dict of heartbeat>],}
    e.g.,
    {'cluster': [{'heartbeat_mode': 'global',
                  'node_count': '3',
                  'name': 'ocfs2test'}],
     'node': [{'ip_port': '7777',
               'cluster': 'ocfs2test',
               'ip_address': '10.149.76.185',
               'number': '0',
               'name': 'ovs085'},
              {'ip_port': '7777',
               'cluster': 'ocfs2test',
               'ip_address': '10.149.76.186',
               'number': '1',
               'name': 'ovs086'},
              {'ip_port': '7777',
               'cluster': 'ocfs2test',
               'ip_address': '10.149.76.187',
               'number': '2',
               'name': 'ovs087'}],
     'heartbeat': [{'cluster': 'ocfs2test',
                    'region': '380832D5B2A64C30948744568F28B302'}}
    """
    conf_dict = {'node': [], 'cluster': [], 'heartbeat': []}
    for line in file_readlines(cfg_file):
        line = line.split('#')[0].strip()
        if not line:
            continue
        elif line.startswith('node:'):
            d = {}
            conf_dict['node'].append(d)
        elif line.startswith('cluster:'):
            d = {}
            conf_dict['cluster'].append(d)
        elif line.startswith('heartbeat:'):
            d = {}
            conf_dict['heartbeat'].append(d)
        else:
            parts = map(string.strip, line.split('='))
            if len(parts) == 2 and parts[0]:
                d[parts[0]] = parts[1]
    return conf_dict


def config_cluster_layout(cluster_conf, cfg_file=CLUSTER_CONF):
    if os.path.exists(cfg_file):
        cluster_layout = parse_cluster_layout(cfg_file)
    else:
        cluster_layout = {'cluster': [], 'node': [], 'heartbeat': []}

    cluster = O2CBCluster(cfg_file)

    for cl in cluster_layout['cluster']:
        if cl not in cluster_conf['cluster']:
            cluster.remove_cluster(cl)
    for cl in cluster_conf['cluster']:
        if cl not in cluster_layout['cluster']:
            cluster.add_cluster(cl)
            cluster.heartbeat_mode(cl)

    for nd in cluster_layout['node']:
        if nd not in cluster_conf['node']:
            cluster.remove_node(nd)
    for nd in cluster_conf['node']:
        if nd not in cluster_layout['node']:
            cluster.add_node(nd)

    #for hb in cluster_layout['heartbeat']:
    #    if hb not in cluster_conf['heartbeat']:
    #        cluster.remove_heartbeat(hb)
    for hb in cluster_conf['heartbeat']:
        if hb not in cluster_layout['heartbeat']:
            cluster.add_heartbeat(hb)


def discover_o2cb_config(doc):
    o2cb_conf = get_o2cb_conf()
    doc.push("O2CB_Config")
    for k, v in o2cb_conf.items():
        doc.add(k, v)
    doc.pop()


def discover_stored_cluster_info(doc):
    node_items = [ ("number", "Number"),
                   ("ip_port", "IP_Port"),
                   ("ip_address", "IP_Address"),
                   ("name", "Name"),
                   ("cluster", "Cluster_Name"), ]

    cluster_items = [ ("name", "Name"),
                      ("node_count", "Node_Count"),
                      ("heartbeat_mode", "Heartbeat_Mode"), ]

    heartbeat_items = [ ("region", "Region"),
                        ("cluster", "Cluster"), ]

    doc.push("Stored")

    parsed_config = parse_cluster_layout()

    doc.push("Clusters")

    cluster_list = parsed_config["cluster"]
    if not cluster_list:
        raise Exception("cluster.conf 'cluster:' section is invalid or missing")

    for cluster in cluster_list:
        doc.push("Cluster")
        for (file_kw, element) in cluster_items:
            if file_kw in cluster:
                doc.add(element, cluster[file_kw])
            else:
                raise Exception("cluster.conf missing '%s' line" % file_kw)
        doc.pop() # Cluster

    doc.pop() # Clusters

    doc.push("Heartbeats")

    heartbeat_list = parsed_config["heartbeat"]
    if not heartbeat_list:
        raise Exception("cluster.conf 'heartbeat:' section is invalid or missing")

    for heartbeat in heartbeat_list:
        doc.push("Heartbeat")
        for (file_kw, element) in heartbeat_items:
            if file_kw in heartbeat:
                doc.add(element, heartbeat[file_kw])
            else:
                raise Exception("cluster.conf missing '%s' line" % file_kw)
        doc.pop() # Heartbeat

    doc.pop() # Heartbeats

    node_list = parsed_config["node"]
    if not node_list:
        raise Exception("cluster.conf 'node:' section is invalid or missing")

    doc.push("Nodes")

    for node in node_list:
        doc.push("Node")
        for (file_kw, element) in node_items:
            if file_kw in node:
                doc.add(element, node[file_kw])
            else:
                raise Exception("cluster.conf missing '%s' line for node %s" % (file_kw, node))
        doc.pop() # Node

    doc.pop() # Nodes

    doc.pop() # Stored


def get_active_cluster_info(cluster_name, registered_hostname):
    mybase = os.path.join(CLUSTER_CONFIGFS, cluster_name,
                          "node", registered_hostname)
    if not os.path.isdir(mybase):
        return (None, None, None)

    fp = open(os.path.join(mybase, "local"), "r")
    if fp.readline().strip() != "1":
        raise Exception("Cluster sysfs entry for this node not marked local")
    fp.close()

    fp = open(os.path.join(mybase, "num"), "r")
    num = fp.readline().strip()
    if not num.isdigit():
        raise Exception("Cluster sysfs node number not a number")
    fp.close()

    fp = open(os.path.join(mybase, "ipv4_address"), "r")
    ip_addr = fp.readline().strip()
    if not ip_addr:
        raise Exception("Cluster sysfs IP address missing")
    fp.close()

    fp = open(os.path.join(mybase, "ipv4_port"), "r")
    ip_port = fp.readline().strip()
    if not ip_port:
        raise Exception("Cluster sysfs IP port missing")
    fp.close()

    return (num, ip_addr, ip_port)


def cleanup_cluster_conf():

    logging.debug("Begin cluster cleanup")

    # 1 - remove cluster.conf
    if os.path.exists(CLUSTER_CONF):
        try:
            os.remove(CLUSTER_CONF)
        except Exception, err1:
            logging.error("Failed to remove cluster config file - error %s", err1)
    else:
        logging.error("Error - cluster.conf does not exist")

   #2 - clear o2cb.conf by removing currently stamped cluster
    clean_conf = {
        "O2CB_ENABLED": "false",
        "O2CB_STACK": "o2cb",
        "O2CB_BOOTCLUSTER": "",
        "O2CB_HEARTBEAT_THRESHOLD": "31",
        "O2CB_IDLE_TIMEOUT_MS": "60000",
        "O2CB_KEEPALIVE_DELAY_MS": "2000",
        "O2CB_RECONNECT_DELAY_MS": "2000",
        }
    try:
        set_o2cb_conf(clean_conf)
    except Exception, err2:
        logging.error("Failed to clear O2CB config file - error %s", err2)

    #3 - offline cluster
    try:
         server_cluster_offline()
    except Exception, err3:
         logging.error("Failed to offline cluster - error %s", err3)

    #4 - clean up DB record - remove cluster pool FS related entries
    clean_poolfs_info = {
         "poolfs_type": None,
         "poolfs_target": None,
         "poolfs_uuid": None,
         "poolfs_nfsbase_uuid": None,
         }
    try:
        update_db(LOCAL_SERVER_DB, clean_poolfs_info)
    except Exception, err4:
        logging.error("Failed to update local server db - error %s", err4)

    logging.debug("End cluster cleanup")
    #end cleanup procedure

