# Copyright (C) 2006-2014 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import shutil

from agent.api.base import Agent
from agent.lib import filelock, assembly
from agent.lib.db import delete_item, write_item
from agent.lib.template import (create_template, delete_template,
                                get_template_path, get_template_config,
                                configure_template, unconfigure_template,
                                set_template_config, import_template,
                                import_vm)
from agent.lib.settings import (get_vm_lock_file, get_vm_path,
                                get_cluster_db_home, format_uuid)
from agent.lib.xenvm import (create_cfg, get_cfg_file, get_cfg_path,
                             read_cfg, write_cfg)


class Hypervisor(Agent):
    """
    Class for hypervisor management.

    """

    def create_vm(self, repo_id, vm_id, params):
        """
        Create a VM.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @param params: VM configuration parameters
        @type params: C{dict}

        @raise IModuleException: on failure

        """
        lock = filelock.FileLock(get_vm_lock_file(repo_id, vm_id))
        lock.acquire(wait=120)
        try:
            path = get_vm_path(repo_id, vm_id)
            if not os.path.exists(path):
                os.makedirs(path)
            try:
                create_cfg(get_cfg_file(path), params)
            except Exception:
                shutil.rmtree(path)
                raise
        finally:
            lock.release()


    def delete_vm(self, repo_id, vm_id):
        """
        Delete a VM.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @raise IModuleException: on failure

        """
        lock = filelock.FileLock(get_vm_lock_file(repo_id, vm_id))
        lock.acquire(wait=120)
        try:
            shutil.rmtree(get_vm_path(repo_id, vm_id))
        finally:
            lock.release()


    def start_vm(self, repo_id, vm_id):
        """
        Start a VM.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def stop_vm(self, repo_id, vm_id, force=False):
        """
        Stop a VM.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @param force: Forcibly stop the VM.
        @type force: C{bool}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def reboot_vm(self, repo_id, vm_id, wait=True):
        """
        Reboot a VM.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @param wait: Wait for reboot to complete
        @type wait: C{bool}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def pause_vm(self, repo_id, vm_id):
        """
        Pause a VM.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def unpause_vm(self, repo_id, vm_id):
        """
        Unpause a VM.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def save_vm(self, repo_id, vm_id, checkpoint=False):
        """
        Save a VM's running state to disk and suspend it.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @param checkpoint: Checkpoint to disk and continue running?
        @type checkpoint: C{bool}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def restore_vm(self, repo_id, vm_id, paused=False):
        """
        Restore a VM's running state from disk and resume it.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @param paused: Paused?
        @type paused: C{bool}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def migrate_vm(self, repo_id, vm_id, dest, live=True, ssl=True):
        """
        Migrate a VM.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @param dest: Destination host
        @type dest: C{str}

        @param live: Live migrate the VM?
        @type live: C{bool}

        @param ssl: Use SSL to perform the migration?
        @type ssl: C{bool}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def trigger_vm(self, repo_id, vm_id, name, vcpu=None):
        """
        Issue a platform-specific trigger to VM.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @param name: Name of the trigger (nmi|reset|init|s3resume|power)
        @type name: C{str}

        @param vcpu: VCPU to which to issue the trigger
        @type vcpu: C{str}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def sysrq_vm(self, repo_id, vm_id, letter):
        """
        Issue a platform-specific system-request a VM.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @param letter: Letter of the sysrq to issue.
        @type letter: C{str}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def dump_vm_core(self, repo_id, vm_id, live=False, crash=False,
                     reset=False):
        """
        Create a VM coredump file.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @param live: Live
        @type live: C{bool}

        @param crash: Crash
        @type crash: C{bool}

        @param reset: Reset
        @type reset: C{bool}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def list_vm_core(self, repo_id, vm_id):
        """
        List the VM coredumps for a VM.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def delete_vm_core(self, repo_id, vm_id, core_date):
        """
        Delete a VM coredump.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @param core_date: Date of coredump file
        @type core_date: C{str}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def configure_vm(self, repo_id, vm_id, params):
        """
        Configure a VM.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @param params: VM configuration params
        @type params: C{dict}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def unconfigure_vm(self, repo_id, vm_id, params):
        """
        Unconfigure a VM.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @param params: VM configuration params
        @type params: C{dict}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()


    def get_vm_config(self, repo_id, vm_id):
        """
        Get a VM's configuration.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @raise IModuleException: on failure

        """
        lock = filelock.FileLock(get_vm_lock_file(repo_id, vm_id))
        lock.acquire(wait=120)
        try:
            return read_cfg(get_cfg_path(repo_id, vm_id))
        finally:
            lock.release()


    def set_vm_config(self, repo_id, vm_id, params):
        """
        Set a VM's configuration.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @param params: VM configuration params
        @type params: C{dict}

        @raise IModuleException: on failure

        """
        lock = filelock.FileLock(get_vm_lock_file(repo_id, vm_id))
        lock.acquire(wait=120)
        try:
            return write_cfg(get_cfg_path(repo_id, vm_id), params)
        finally:
            lock.release()


    def list_vm(self, repo_id, vm_id):
        """
        List a VM.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def list_vms(self):
        """
        List all VMs.

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def send_to_guest(self, repo_id, vm_id, params):
        """
        Send VMAPI key-value parameters to VM.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @param params: VMAPI key-value parameters
        @type params: C{dict}

        @raise IModuleException: on failure
        """
        raise NotImplementedError()

    def configure_vm_ha(self, repo_id, vm_id, enable_ha=True):
        """
        Configure a VM for HA.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @raise IModuleException: on failure
        """
        path = get_cfg_path(repo_id, vm_id)
        uuid = format_uuid(read_cfg(path)['uuid'])
        db_home = get_cluster_db_home()
        if enable_ha:
            d = {'repo_id': repo_id, 'vm_id': vm_id}
            write_item('monitored_vms', uuid, d, db_home)
        else:
            delete_item('monitored_vms', uuid, db_home)


    def setup_migration_target(self, repo_id, vm_id):
        """
        Setup the server as a migration target for the indicated VM.
        This is function is invoked before the VM is migrated to this
        server.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def cleanup_migration_target(self, repo_id, vm_id):
        """
        Cleanup the server that was configured as a migration target
        for the indicated VM. This is function is invoked to cleanup
        any migration setting when the migration to this server has
        failed.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: VM uuid
        @type vm_id: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()

    #
    # Template
    #

    def create_template(self, repo_id, template_id, params):
        """
        Create a template.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param template_id: Template uuid
        @type template_id: C{str}

        @raise IModuleException: on failure

        """
        return create_template(repo_id, template_id, params)


    def delete_template(self, repo_id, template_id):
        """
        Delete a template.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: Template uuid
        @type vm_id: C{str}

        @raise IModuleException: on failure

        """
        return delete_template(repo_id, template_id)


    def import_vm(self, repo_id, vm_id, url_list, option=None):
        """
        Import a vm from one or more URLs.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param vm_id: Template uuid
        @type vm_id: C{str}

        @param url_list: List of URLs from which to download and import VMs
        @type url_list: C{list}

        @param option: Special non-default options for the download
        @type option: C{dict}

        @raise IModuleException: on failure

        """
        if not url_list:
            raise RuntimeError('No URL is provided.')
        return import_vm(repo_id, vm_id, url_list, option)


    def import_template(self, repo_id, template_id, url_list, option=None):
        """
        Import a template from one or more URLs.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param template_id: Template uuid
        @type template_id: C{str}

        @param url_list: List of URLs from which to download and import templates
        @type url_list: C{list}

        @param option: Special non-default options for the download
        @type option: C{dict}

        @raise IModuleException: on failure

        """
        if not url_list:
            raise RuntimeError('No URL is provided.')
        return import_template(repo_id, template_id, url_list, option)


    def configure_template(self, repo_id, template_id, params):
        """
        Configure a template.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param template_id: Template uuid
        @type template_id: C{str}

        @param params: Template configuration parameters
        @type params: C{dict}

        @raise IModuleException: on failure

        """
        lock = filelock.FileLock(get_vm_lock_file(repo_id, template_id))
        lock.acquire(wait=120)
        try:
            return configure_template(get_template_path(repo_id, template_id),
                                      params)
        finally:
            lock.release()


    def unconfigure_template(self, repo_id, template_id, params):
        """
        Unconfigure a template.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param template_id: Template uuid
        @type template_id: C{str}

        @param params: Template configuration parameters
        @type params: C{dict}

        @raise IModuleException: on failure
        """
        lock = filelock.FileLock(get_vm_lock_file(repo_id, template_id))
        lock.acquire(wait=120)
        try:
            return unconfigure_template(get_template_path(repo_id, template_id),
                                        params)
        finally:
            lock.release()


    def get_template_config(self, repo_id, template_id):
        """
        Get configuration parameters for a template.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param template_id: Template uuid
        @type template_id: C{str}

        @raise IModuleException: on failure
        """
        lock = filelock.FileLock(get_vm_lock_file(repo_id, template_id))
        lock.acquire(wait=120)
        try:
            return get_template_config(get_template_path(repo_id, template_id))
        finally:
            lock.release()


    def set_template_config(self, repo_id, template_id, params):
        """
        Set configuration parameters for a template.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param template_id: Template uuid
        @type template_id: C{str}

        @param params: Template configuration parameters
        @type params: C{dict}

        @raise IModuleException: on failure

        """
        lock = filelock.FileLock(get_vm_lock_file(repo_id, template_id))
        lock.acquire(wait=120)
        try:
            return set_template_config(get_template_path(repo_id, template_id),
                                       params)
        finally:
            lock.release()

    #
    # Assembly
    #

    def import_assembly(self, repo_id, assembly_id, url, option=None):
        """
        Import an assembly.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param assembly_id: Assembly uuid
        @type assembly_id: C{str}

        @param url: URL from which to download the assembly
        @type url: C{str}

        @param option: Special non-default options for the download
        @type option: C{dict}

        @raise IModuleException: on failure

        """
        return assembly.import_assembly(repo_id, assembly_id, url, option)


    def deploy_assembly(self, repo_id, assembly_id, to_deploy=None,
                        target_repo_id=None, option=None):
        """
        Deploy an assembly to a repository.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param assembly_id: Assembly uuid
        @type assembly_id: C{str}

        @param to_deploy: Specific VM to deploy
        @type to_deploy: C{str}

        @param target_repo_id: Target repository UUID
        @type target_repo_id: C{str}

        @param option: Special non-default options for the download
        @type option: C{dict}

        @raise IModuleException: on failure

        """
        return assembly.deploy_assembly(repo_id, assembly_id, to_deploy,
                                        target_repo_id, option)


    def delete_assembly(self, repo_id, assembly_id):
        """
        Delete an assembly.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param assembly_id: Assembly uuid
        @type assembly_id: C{str}

        @raise IModuleException: on failure

        """
        return assembly.delete_assembly(repo_id, assembly_id)


    def get_assembly_config(self, repo_id, assembly_id):
        """
        Get the configuration for an assembly.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param assembly_id: Assembly uuid
        @type assembly_id: C{str}

        @raise IModuleException: on failure

        """
        return assembly.get_assembly_config(repo_id, assembly_id)


    def get_assembly_config_xml(self, repo_id, assembly_id):
        """
        Get the OVF XML data for an assembly.

        @param repo_id: Repository uuid
        @type  repo_id: C{str}

        @param assembly_id: Assembly uuid
        @type assembly_id: C{str}

        @raise IModuleException: on failure

        """
        return assembly.get_assembly_config_xml(repo_id, assembly_id)


    def set_assembly_config(self, repo_id, assembly_id, cfg):
        return assembly.set_assembly_config(repo_id, assembly_id, cfg)


    def set_assembly_config_xml(self, repo_id, assembly_id, cfg):
        return assembly.set_assembly_config_xml(repo_id, assembly_id, cfg)


    def create_assembly(self, repo_id, assembly_id, templates):
        return assembly.create_assembly(repo_id, assembly_id, templates)


    def pack_assembly(self, repo_id, assembly_id):
        return assembly.pack_assembly(repo_id, assembly_id)


    def unpack_assembly(self, repo_id, assembly_id):
        return assembly.unpack_assembly(repo_id, assembly_id)


    def assembly_add_file(self, repo_id, assembly_id, url,
                          filename, option=None):
        return assembly.assembly_add_file(repo_id, assembly_id, url,
                                          filename, option)


    def assembly_del_file(self, repo_id, assembly_id, filename):
        return assembly.assembly_del_file(repo_id, assembly_id, filename)


    def assembly_rename_file(self, repo_id, assembly_id, filename,
                             new_filename):
        return assembly.assembly_rename_file(repo_id, assembly_id, filename,
                                             new_filename)
