# Copyright (C) 2006-2014 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import sys
import signal
import socket
import time
import logging
import traceback
import agent.lib.constants
from SocketServer import ForkingMixIn
from agent.utils import xmlrpclib
from agent.utils.SimpleXMLRPCServer import (SimpleXMLRPCServer,
        SimpleXMLRPCRequestHandler, SimpleXMLRPCDispatcher)
from OpenSSL import SSL
from OpenSSL.crypto import FILETYPE_PEM, load_certificate, X509_verify_cert_error_string
from agent.lib.daemon import (ignore_eintr, init_log, reload_xmlrpc_server,
        stop_process, clean_up_lock)
from agent.lib.constants import (SSL_SERVER_PRIVATE_KEY,
                                 SSL_SERVER_CERTIFICATE,
                                 SSL_SIGNED_SERVER_CERTIFICATE,
                                 SSL_MANAGER_CERTIFICATE)
from agent.lib.config import get_config
from agent.lib.password import encrypt_password, read_password, update_password
from agent.lib.settings import get_api_version, get_pool_member_ip_list
from agent.lib.service import get_api
from agent.target.host import get_ovs_info


PASSWORD = None
SERVER = None
SSL_AUTH_OK = None


def get_passwd():
    global PASSWORD
    if PASSWORD is None:
        PASSWORD = read_password()
    return PASSWORD


def sig_handler(sig, frame):
    if sig == signal.SIGHUP:
        global PASSWORD
        PASSWORD = None
    if sig == signal.SIGUSR1 and SERVER:
        SERVER.stop = True
        # wait for take_ownership()/release_ownership() to finish
        time.sleep(0.5)
        SERVER.server_close()
        stop_process(os.getpid(), childrenonly=True)
        # wait for all children to terminate
        while SERVER.active_children:
            SERVER.collect_children()
        clean_up_lock()


def ssl_verify(conn, cert, errno, depth, ok):
    global SSL_AUTH_OK
    if not ok:
        logging.error('SSL verification failed: errno: %s (%s), depth: %s, connection: %s',
                      errno, X509_verify_cert_error_string(errno), depth, conn.getpeername())
        logging.error('Certificate subject: %s, issuer: %s, valid not before: '
                      '%s, valid not after: %s', str(cert.get_subject()),
                      str(cert.get_issuer()), str(cert.get_notBefore()),
                      str(cert.get_notAfter()))
        if cert.has_expired():
            logging.error('Certificate expired: check cert valid time and system time.')
        SSL_AUTH_OK = False
    else:
        if SSL_AUTH_OK is None:
            SSL_AUTH_OK = True
    return 1


def auth(username, passwd):
    if SSL_AUTH_OK:
        return
    password_dict = get_passwd()
    method, stored_passwd = password_dict.get(username, ('CRYPT', '!'))
    if stored_passwd != encrypt_password(passwd, method, stored_passwd):
        raise Exception('Authorization Failed')
    if method == 'SHA':
        logging.warn('Replacing SHA password for user: %s', username)
        update_password(username, passwd)
        reload_xmlrpc_server()


def auth_ip(ip):
    if SSL_AUTH_OK:
        return
    if ip != "127.0.0.1" and ip not in get_pool_member_ip_list():
        raise Exception('Authorization failed: server not in the server pool: %s' % ip)


class AgentXMLRPCRequestHandler(SimpleXMLRPCRequestHandler):
    def setup(self):
        config = get_config()
        if config.get("security", "ssl").lower().startswith("enable"):
            self.connection = self.request
            self.rfile = socket._fileobject(self.request, "rb", self.rbufsize)
            self.wfile = socket._fileobject(self.request, "wb", self.wbufsize)
        else:
            SimpleXMLRPCRequestHandler.setup(self)

    def do_POST(self):
        # Check that the path is legal
        if not self.is_rpc_path_valid():
            self.report_404()
            return

        try:
            src_ip = self.client_address[0]
            agent.lib.constants.MANAGER_SRC_IP = src_ip
            if "authorization" in self.headers:
                username, password = self.headers["authorization"].split(" ")[1].decode("base64").split(":")
                auth(username, password)
            else:
                auth_ip(src_ip)
        except:
            errmsg = "Unauthorized access attempt from %s!" % (self.client_address,)
            logging.exception(errmsg)
            self.send_error(403, errmsg)
            return

        config = get_config()
        if config.get("security", "ssl").lower().startswith("enable"):
            try:
                # Get arguments by reading body of request.
                # We read this in chunks to avoid straining
                # socket.read(); around the 10 or 15Mb mark, some platforms
                # begin to have problems (bug #792570).
                max_chunk_size = 10*1024*1024
                size_remaining = int(self.headers["content-length"])
                L = []
                while size_remaining:
                    chunk_size = min(size_remaining, max_chunk_size)
                    chunk = self.rfile.read(chunk_size)
                    if not chunk:
                        break
                    L.append(chunk)
                    size_remaining -= len(L[-1])
                data = ''.join(L)
    
                # In previous versions of SimpleXMLRPCServer, _dispatch
                # could be overridden in this class, instead of in
                # SimpleXMLRPCDispatcher. To maintain backwards compatibility,
                # check to see if a subclass implements _dispatch and dispatch
                # using that method if present.
                response = self.server._marshaled_dispatch(
                        data, getattr(self, '_dispatch', None), self.path
                    )
            except Exception, e: # This should only happen if the module is buggy
                # internal error, report as HTTP server error
                self.send_response(500)
    
                # Send information about the exception if requested
                if hasattr(self.server, '_send_traceback_header') and \
                        self.server._send_traceback_header:
                    self.send_header("X-exception", str(e))
                    self.send_header("X-traceback", traceback.format_exc())
    
                self.end_headers()
            else:
                # got a valid XML RPC response
                self.send_response(200)
                self.send_header("Content-type", "text/xml")
                self.send_header("Content-length", str(len(response)))
                self.end_headers()
                self.wfile.write(response)
    
                # shut down the connection
                self.wfile.flush()
                self.connection.shutdown()
        else:
            try:
                # Get arguments by reading body of request.
                # We read this in chunks to avoid straining
                # socket.read(); around the 10 or 15Mb mark, some platforms
                # begin to have problems (bug #792570).
                max_chunk_size = 10*1024*1024
                size_remaining = int(self.headers["content-length"])
                L = []
                while size_remaining:
                    chunk_size = min(size_remaining, max_chunk_size)
                    chunk = self.rfile.read(chunk_size)
                    if not chunk:
                        break
                    L.append(chunk)
                    size_remaining -= len(L[-1])
                data = ''.join(L)
    
                # In previous versions of SimpleXMLRPCServer, _dispatch
                # could be overridden in this class, instead of in
                # SimpleXMLRPCDispatcher. To maintain backwards compatibility,
                # check to see if a subclass implements _dispatch and dispatch
                # using that method if present.
                response = self.server._marshaled_dispatch(
                        data, getattr(self, '_dispatch', None), self.path
                    )
            except Exception, e: # This should only happen if the module is buggy
                # internal error, report as HTTP server error
                self.send_response(500)
    
                # Send information about the exception if requested
                if hasattr(self.server, '_send_traceback_header') and \
                        self.server._send_traceback_header:
                    self.send_header("X-exception", str(e))
                    self.send_header("X-traceback", traceback.format_exc())
    
                self.end_headers()
            else:
                # got a valid XML RPC response
                self.send_response(200)
                self.send_header("Content-type", "text/xml")
                self.send_header("Content-length", str(len(response)))
                self.end_headers()
                self.wfile.write(response)
    
                # shut down the connection
                self.wfile.flush()
                self.connection.shutdown(1)

    def log_message(self, format, *args):
        logging.info(format % args)


class OVSAgentServer(ForkingMixIn, SimpleXMLRPCServer):

    allow_reuse_address = True
    max_children = 500 # default 40 in ForkingMixIn

    def __init__(self, addr, requestHandler=AgentXMLRPCRequestHandler):
        SimpleXMLRPCServer.__init__(self, addr, requestHandler, logRequests=False, allow_none=True)
        self.stop = False
        config = get_config()
        if config.get("security", "ssl").lower().startswith("enable"):
            certfile = SSL_SERVER_CERTIFICATE
            if os.path.exists(SSL_SIGNED_SERVER_CERTIFICATE):
                certfile = SSL_SIGNED_SERVER_CERTIFICATE
            ctx = SSL.Context(SSL.SSLv23_METHOD)
            ctx.set_options(SSL.OP_NO_SSLv2)
            ctx.set_options(SSL.OP_NO_SSLv3)
            ctx.set_cipher_list('DEFAULT:!aNULL:!AHD:!eNULL:!LOW:!MEDIUM:!EXP')
            ctx.use_privatekey_file(SSL_SERVER_PRIVATE_KEY)
            ctx.use_certificate_file(certfile)

            if os.path.exists(SSL_MANAGER_CERTIFICATE):
                ctx.load_verify_locations(SSL_MANAGER_CERTIFICATE)
                ctx.set_verify(SSL.VERIFY_PEER, ssl_verify)
                manager_certfile = open(SSL_MANAGER_CERTIFICATE)
                try:
                    manager_cert = manager_certfile.read()
                finally:
                    manager_certfile.close()
                manager_cert = manager_cert.strip()
                if manager_cert:
                    cacert = load_certificate(FILETYPE_PEM, manager_cert)
                    caissuer = cacert.get_issuer()
                    ctx.set_client_ca_list([caissuer])
                    casubject = cacert.get_subject()
                    logging.debug('Manager cert subject: %s', casubject.get_components())
                    logging.debug('Manager cert issuer: %s', caissuer.get_components())

            self.socket = SSL.Connection(ctx, socket.socket(self.address_family, self.socket_type))
            self.server_bind()
            self.server_activate()
        self.dispatchers = {}
        # /api/0 is only a workaround to fix released Manager 3.1.1 bug: it
        # will pass down '/api/0' if no matching version. Manager will use "/"
        # in 3.2.1+.
        # /api/1 is another workaround to fix released Manager 3.1.1 bug: it
        # is hardcoded to pass down '/api/1' for get_api_version().
        # Manager will use "/" in 3.2.1+.
        requestHandler.rpc_paths = ['/', '/RPC2', '/api/0', '/api/1']
        api_version = get_api_version()
        for version in api_version:
            api = get_api(version)
            dispatcher = SimpleXMLRPCDispatcher(allow_none=True)
            for (api, func) in api.items():
                dispatcher.register_function(func, api)
            rpc_path = '/api/%s' % version
            self.dispatchers[rpc_path] = dispatcher
            if rpc_path not in requestHandler.rpc_paths:
                requestHandler.rpc_paths.append(rpc_path)

        latest_api_rpc_path = '/api/%s' % api_version[-1]
        self.dispatchers['/'] = self.dispatchers[latest_api_rpc_path]
        self.dispatchers['/RPC2'] = self.dispatchers[latest_api_rpc_path]
        self.dispatchers['/api/0'] = self.dispatchers[latest_api_rpc_path]
        self.dispatchers['/api/1'] = self.dispatchers[latest_api_rpc_path]

    def _marshaled_dispatch(self, data, dispatch_method=None, path=None):
        try:
            response = self.dispatchers[path]._marshaled_dispatch(
                data, dispatch_method)
        except:
            # report low level exception back to server
            # (each dispatcher should have handled their own
            # exceptions)
            response = xmlrpclib.dumps(
                xmlrpclib.Fault(1, "%s:%s" % (sys.exc_type, sys.exc_value)),
                allow_none=True)
        return response

    def verify_request(self, request, client_address):
        """
        overwrite original verify_request to implement IP based access control
        """
        #--------------------<> internal function <>----
        def match(rule, ip):
            #check if ip is in the range
            try:
                #e.g.: 219.142.73.80
                if ip == rule:
                    return True
                #e.g.: 219.142.73.*
                if rule.endswith("*") and ip.startswith(rule[:-1]):
                    return True
                #e.g.: 219.142.73.0/20
                vip, bits = rule.split("/")
                vip_str = "".join(["%02x" % int(x) for x in vip.split(".")])
                ip_str = "".join(["%02x" % int(x) for x in ip.split(".")])
                q, r = divmod(int(bits), 4)
                if vip_str[:q] == ip_str[:q]:
                    if int(vip_str[q], 16) >> (4 - r) == int(ip_str[q], 16) >> (4 - r):
                        return True
            except: #in case there are errors in rules
                pass
        #--------------------<> main processing <>----
        #check ip access control rules, sequence: allow=>deny
        (ip, port) = client_address

        # always allow connect from localhost
        if ip == '127.0.0.1':
            return True

        config = get_config()
        #allow rules
        allow = False
        allow_rules = config.get("security", "allow").split(",")
        for rule in allow_rules:
            if match(rule.strip(), ip):
                allow = True
                break
        if not allow:
            return False

        #deny rules
        deny = False
        deny_rules = config.get("security", "deny").split(",")
        for rule in deny_rules:
            if match(rule.strip(), ip):
                deny = True
                break
        if deny:
            return False

        #nothing wrong, finally
        return True

    def serve_forever(self):
        while not self.stop:
            self.handle_request()


def serve_forever():
    signal.signal(signal.SIGHUP, sig_handler)
    signal.signal(signal.SIGUSR1, sig_handler)
    hostname = ""
    hostip = ""
    try:
        hostname = socket.gethostname()
        hostip = socket.gethostbyname(hostname)
    except Exception, err:
        logging.warn("Error getting hostname and IP: %s", err)
    logging.info("Oracle VM Server version: %s, hostname: %s, ip: %s",
                 get_ovs_info(), hostname, hostip)
    config = get_config()
    global SERVER
    while True:
        try:
            SERVER = OVSAgentServer(('', int(config.get("agent", "port"))))
        except Exception, err:
            logging.debug('Error init XMLRPC Server: %s', err)
            time.sleep(1)
            continue
        logging.info("Oracle VM Agent XMLRPC Server started.")
        try:
            ignore_eintr(OVSAgentServer.serve_forever, SERVER)
        except Exception, err:
            logging.info('Error reloading XMLRPC Server: %s', err)
            # make sure the server socket is closed
            SERVER.server_close()


if __name__ == '__main__':
    init_log()
    serve_forever()
