# Copyright (C) 2006-2016 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import os
import re
import time
import xen.lowlevel.xc
import xen.xend.XendClient
from xen.xend.xenstore.xstransact import xstransact
from agent.target.host import run_cmd, get_server_unique_id
from agent.target.hypervisor import get_vms
from agent.lib.stats import StatsCollectorBase
from agent.lib.notification import send_statistic

MAX_LONG = 2**63 - 1

# System and VM stats sent to manager could be unsigned long values which may 
# exceed the range of Java long type. Round the values to length (2^63 - 1)
def convert_to_long(number):
    return long(number) % (MAX_LONG)

def read_file(*args):
    tmpfile = open(os.path.join(*args))
    try:
        return tmpfile.read().strip()
    finally:
        tmpfile.close()


def read_xenstore(path):
    try:
        return xstransact.Read(path).strip()
    except Exception, err:
        logging.debug('Error reading xenstore (%s): %s', path, err)


def get_vm_stats():
    try:
        return get_vms()
    except Exception, err:
        logging.debug('Error getting VM stats: %s', err)


def get_disk_uuid(dev):
    try:
        info = run_cmd(['dmsetup', 'info', '--columns', '--noheadings',
                        '--options=UUID', '/dev/%s' % dev])
        if not info.startswith('part'):
            return info.rstrip('\n').replace('mpath-', '')
    except Exception:
        pass


def get_disk_stat():
    disk_read = {}
    disk_write = {}
    try:
        f = open('/proc/diskstats')
        try:
            for line in f:
                stat = line.split()
                if not stat[2].startswith('dm-'):
                    continue
                disk_uuid = get_disk_uuid(stat[2])
                if not disk_uuid:
                    continue
                read = long(stat[5]) << 9
                write = long(stat[9]) << 9
                disk_read[disk_uuid] = convert_to_long(read)
                disk_write[disk_uuid] = convert_to_long(write)
        finally:
            f.close()
    except Exception, err:
        logging.debug('Error getting Server disk read/write: %s', err)
    return disk_read, disk_write


def get_port_stat():
    port_rx = {}
    port_tx = {}
    try:
        basedir = '/sys/class/net'
        for entry in os.listdir(basedir):
            statdir = os.path.join(basedir, entry, 'statistics')
            if not os.path.exists(statdir):
                continue
            if entry.startswith('vif') or entry == 'lo':
                continue
            rx = read_file(statdir, 'rx_bytes')
            tx = read_file(statdir, 'tx_bytes')
            port_rx[entry] = convert_to_long(rx)
            port_tx[entry] = convert_to_long(tx)
    except Exception, err:
        logging.debug('Error getting Server port RX/TX: %s', err)
    return port_rx, port_tx


def get_per_cpu_idle_time(max_cpus):
    # cpu idletime from xc.getcpuinfo() is in nanosecond. Convert it into second.
    per_cpu_idle_time = []
    try:
        xc = xen.lowlevel.xc.xc()
        cpu_info = xc.getcpuinfo(max_cpus)
        per_cpu_idle_time = [float(x['idletime']) / (10 ** 9) for x in cpu_info]
    except Exception, err:
        logging.debug('Error getting Server per physical CPU idle time: %s', err)
    return per_cpu_idle_time


def get_vm_disk_stat(domid):
    disk_read = {}
    disk_write = {}
    try:
        basedir = '/sys/bus/xen-backend/devices'
        for entry in os.listdir(basedir):
            if not re.match(r'^vbd-%s-\d+$' % domid, entry):
                continue
            nodename = read_file(basedir, entry, 'nodename')
            backend = read_xenstore('/local/domain/0/%s/params' % nodename)
            if not backend:
                continue
            # remove leading 'aio:', 'ioemu:', etc.
            backend = backend.split(':', 1)[-1]
            read = read_file(basedir, entry, 'statistics/rd_sect')
            write = read_file(basedir, entry, 'statistics/wr_sect')
            disk_read[backend] = convert_to_long(read)
            disk_write[backend] = convert_to_long(write)
    except Exception, err:
        logging.debug('Error getting VM disk read/write: %s', err)
    return disk_read, disk_write


def get_vm_port_stat(domid):
    port_rx = {}
    port_tx = {}
    try:
        basedir = '/sys/class/net'
        for entry in os.listdir(basedir):
            match = re.match(r'^vif%s\.(?P<index>\d+)$' % domid, entry)
            if not match:
                continue
            index = match.group('index')
            statdir = os.path.join(basedir, entry, 'statistics')
            if not os.path.exists(statdir):
                continue
            # backend TX == VM frontend RX
            # backend RX == VM frontend TX
            rx = read_file(statdir, 'tx_bytes')
            tx = read_file(statdir, 'rx_bytes')
            port_rx[index] = convert_to_long(rx)
            port_tx[index] = convert_to_long(tx)
    except Exception, err:
        logging.debug('Error getting VM port RX/TX: %s', err)
    return port_rx, port_tx


def tmem_parse(line):
    parsed = {}
    for item in line[2:].split(','):
        key, value = item.split(':')
        if key in ['U0', 'U1']:
            parsed[key] = int(value, 16)
        elif key == 'PT':
            parsed[key] = value
        else:
            parsed[key] = int(value)
    return parsed


def tmem_list():
    xc = xen.lowlevel.xc.xc()
    raw = xc.tmem_control(-1, 4, -1, 32768, 1, 0, '')
    if raw == -1:
        return {}
    parsed = {'global': {},
              'time_stats': {},
              'clients': [],
              'pools': [],
              'shared_pools': []}
    for line in raw.splitlines():
        if line.startswith('G='):
            parsed['global'] = tmem_parse(line)
        if line.startswith('T='):
            parsed['time_stats'] = tmem_parse(line)
        if line.startswith('C='):
            parsed['clients'].append(tmem_parse(line))
        if line.startswith('P='):
            parsed['pools'].append(tmem_parse(line))
        if line.startswith('S='):
            parsed['shared_pools'].append(tmem_parse(line))
    result = {}
    for pool in parsed['pools']:
        client_id = int(pool['CI'])
        if client_id not in result:
            result[client_id] = [0, 0]
        pgp_count = pool['Pc']
        if pool['PT'] == 'EP':
            result[client_id][0] += pgp_count
        if pool['PT'] == 'PP':
            result[client_id][1] += pgp_count
    return result


def tmem_query_freeable_mb():
    xc = xen.lowlevel.xc.xc()
    freeable_mb = xc.tmem_control(-1, 8, -1, 0, 0, 0, '')
    if freeable_mb == -1:
        return 0
    else:
        return freeable_mb


class StatsCollector(StatsCollectorBase):
    def __init__(self):
        StatsCollectorBase.__init__(self)
        self._previous_timestamp = 0.0
        self._previous_cpu_time = {}
        self._previous_per_cpu_idle_time = []

    def _run(self):
        current_timestamp = time.time()
        time_delta = current_timestamp - self._previous_timestamp
        if time_delta <= 0:
            self._previous_timestamp = current_timestamp
            return

        vm_stats = get_vm_stats()
        if not vm_stats:
            return
        nodeinfo = dict(xen.xend.XendClient.server.xend.node.info())
        tmeminfo = tmem_list()
        nr_cpus = nodeinfo['nr_cpus']

        disk_read, disk_write = get_disk_stat()
        port_rx, port_tx = get_port_stat()

        ptmem = sum([x for _, x in tmeminfo.values()])
        etmem = tmem_query_freeable_mb()

        dom0_cpu_time = 0.0
        current_cpu_time = {}
        for vm in vm_stats:
            try:
                domid = int(vm['domid'])
                name = vm['name']
                vcpu_count = int(vm['vcpus'])
                target_mem = vm['memory']

                xc = xen.lowlevel.xc.xc()
                try:
                    xcdominfo = xc.domain_getinfo(domid, 1)
                    current_mem = xcdominfo[0]['mem_kb'] / 1024
                except Exception:
                    current_mem = target_mem

                if domid in tmeminfo:
                    vmetmem = tmeminfo[domid][0]
                    vmptmem = tmeminfo[domid][1]
                else:
                    vmetmem = 0
                    vmptmem = 0

                vm_disk_read, vm_disk_write = get_vm_disk_stat(domid)
                vm_port_rx, vm_port_tx = get_vm_port_stat(domid)

                cpu_util = 0.0
                cpu_load = 0.0
                current_vm_cpu_time = float(vm['cpu_time'])
                previous_vm_cpu_time = self._previous_cpu_time.get(vm['uuid'])
                if previous_vm_cpu_time:
                    cpu_time_delta = abs(current_vm_cpu_time - previous_vm_cpu_time)
                    cpu_time_delta_ceiling = time_delta * int(vm['online_vcpus'])
                    if cpu_time_delta > cpu_time_delta_ceiling:
                        cpu_time_delta = cpu_time_delta_ceiling
                    cpu_util = cpu_time_delta / cpu_time_delta_ceiling

                current_cpu_time[vm['uuid']] = current_vm_cpu_time

                # Bypass Domain-0: dom0 statistics will be reported as 'SERVER' type.
                if domid == 0:
                    dom0_cpu_time = current_vm_cpu_time
                    continue

                vm_stat = {
                    'timestamp': convert_to_long(time.time() * 1000),
                    '@type': 'Vm',
                    'type': 'VM',
                    'id': vm['uuid'],
                    'cpuTime': current_vm_cpu_time,
                    'cpuUtil': cpu_util,
                    'currentMemory': convert_to_long(current_mem),
                    'diskRead': vm_disk_read,
                    'diskWrite': vm_disk_write,
                    'domainId': domid,
                    'ephemeralTranscendentMemory': convert_to_long(vmetmem),
                    'maxMemory': convert_to_long(vm['maxmem']),
                    'networkReceive': vm_port_rx,
                    'networkSend': vm_port_tx,
                    'onlineVcpus': int(vm['online_vcpus']),
                    'persistentTranscendentMemory': convert_to_long(vmptmem),
                    'targetMemory': convert_to_long(target_mem),
                }

                send_statistic(vm_stat)
            except Exception, err:
                logging.warn('Error getting VM statistics: %s', err)
                continue

        per_cpu_util = []
        current_per_cpu_idle_time = get_per_cpu_idle_time(nr_cpus)
        if len(current_per_cpu_idle_time) == len(self._previous_per_cpu_idle_time):
            for index, current_cpu_idle_time in enumerate(current_per_cpu_idle_time):
                previous_cpu_idle_time = self._previous_per_cpu_idle_time[index]
                cpu_idle_time_delta = abs(current_cpu_idle_time - previous_cpu_idle_time)
                if cpu_idle_time_delta >  time_delta:
                    cpu_idle_time_delta = time_delta
                cpu_idle = cpu_idle_time_delta / time_delta
                per_cpu_util.append(1 - cpu_idle)

        if per_cpu_util:
            average_cpu_util = sum(per_cpu_util) / float(len(per_cpu_util))
        else:
            average_cpu_util = 0.0

        total_memory = convert_to_long(nodeinfo['total_memory'])
        free_memory = convert_to_long(nodeinfo['free_memory'])
        used_memory = total_memory - free_memory

        meminfo = dict([item.strip() for item in line.split(':', 1)]
                       for line in read_file('/proc/meminfo').splitlines())
        free_swap = convert_to_long(meminfo['SwapFree'].split()[0])

        server_stat = {
            'timestamp': convert_to_long(time.time() * 1000),
            '@type': 'Server',
            'type': 'SERVER',
            'id': get_server_unique_id(),
            'cpuTime': dom0_cpu_time,
            'cpuUtil': average_cpu_util,
            'perCpuUtil': per_cpu_util,
            'diskRead': disk_read,
            'diskWrite': disk_write,
            'ephemeralTranscendentMemory': convert_to_long(etmem),
            'freeMemory': free_memory,
            'networkReceive': port_rx,
            'networkSend': port_tx,
            'persistentTranscendentMemory': convert_to_long(ptmem),
            'usedMemory': used_memory,
            'freeSwap': free_swap,
        }

        send_statistic(server_stat)

        self._previous_timestamp = current_timestamp
        self._previous_cpu_time = current_cpu_time
        self._previous_per_cpu_idle_time = current_per_cpu_idle_time
