# Copyright (C) 2006-2014 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import errno
import os
import string

from agent.lib.config import get_config
from agent.lib.constants import (LOCAL_SERVER_DB,
                                 MEMBERSHIP_STATE_CLUSTERED,
                                 MEMBERSHIP_STATE_HB_CONFIGURED,
                                 MEMBERSHIP_STATE_POOLED,
                                 MEMBERSHIP_STATE_OWNED,
                                 MEMBERSHIP_STATE_UNOWNED,
                                 CLUSTER_STATE_DLM_READY,
                                 SSL_SERVER_CERTIFICATE,
                                 SSL_MANAGER_CERTIFICATE,
                                 SSL_SIGNED_SERVER_CERTIFICATE,
                                 OVS_REPO_HEAD)
from agent.lib.db import read_item, get_cluster_db_home
from agent.target.info import API_VERSION


def get_api_version():
    return API_VERSION


def get_agt_url(host):
    config = get_config()
    if config.get("security", "ssl").lower().startswith("enable"):
        protocol = "https"
    else:
        protocol = "http"
    return "%s://%s:%s" % (protocol, host, config.get("agent", "port"))


def get_repo_path(repo_id):
    return os.path.join(OVS_REPO_HEAD, repo_id)


def get_assembly_path(repo_id, assembly_id):
    return os.path.join(get_repo_path(repo_id), 'Assemblies', assembly_id)


def get_iso_path(repo_id, iso_id):
    return os.path.join(get_repo_path(repo_id), 'ISOs', iso_id)


def get_template_path(repo_id, template_id):
    return os.path.join(get_repo_path(repo_id), 'Templates', template_id)


def get_virtualdisk_path(repo_id, virtualdisk_id):
    return os.path.join(get_repo_path(repo_id), 'VirtualDisks', virtualdisk_id)


def get_vm_path(repo_id, vm_id):
    return os.path.join(get_repo_path(repo_id), 'VirtualMachines', vm_id)


def get_vm_lock_file(repo_id, lock_id):
    return '/var/run/ovs-agent/vm-%s-%s.lock' % (repo_id, lock_id)


def get_cluster_flag():
    return read_item(LOCAL_SERVER_DB, "clustered")


def get_owning_manager_uuid():
    return read_item(LOCAL_SERVER_DB, "manager_uuid")


def get_poolfs_uuid():
    return read_item(LOCAL_SERVER_DB, "poolfs_uuid")


def get_poolfs_target():
    return read_item(LOCAL_SERVER_DB, "poolfs_target")


def get_server_pool_uuid():
    return read_item(LOCAL_SERVER_DB, "pool_uuid")


def get_membership_state():
    """
    Return the current state of the server with respect to
    server pool and cluster membership.

    @return     One of: "Unowned", "Owned", "Pooled",
                "Heartbeat_Configured" and "Clustered".
    """
    if get_cluster_flag():
        return MEMBERSHIP_STATE_CLUSTERED

    if get_poolfs_uuid():
        return MEMBERSHIP_STATE_HB_CONFIGURED

    if get_server_pool_uuid():
        return MEMBERSHIP_STATE_POOLED

    if get_owning_manager_uuid():
        return MEMBERSHIP_STATE_OWNED

    return MEMBERSHIP_STATE_UNOWNED


def get_pool_member_ip_list():
    """
    Get server pool member IP list.

    @return     IP list of server pool members.
    """
    ip_list = []
    state = get_membership_state()

    if state in (MEMBERSHIP_STATE_UNOWNED, MEMBERSHIP_STATE_OWNED):
        ip_list = []
    elif is_clustered():
        ip_list = read_item("server_pool", "pool_member_ip_list",
                            get_cluster_db_home())
    else:
        ip_list = read_item(LOCAL_SERVER_DB, "pool_member_ip_list")

    # None -> []
    return ip_list or []


def is_master():
    """Check whether the current server is the server pool master."""
    my_hostname = read_item(LOCAL_SERVER_DB, "registered_hostname")
    if not my_hostname:
        return False
    master_hostname = read_item("server_pool", "pool_master_hostname",
                                get_cluster_db_home())
    return my_hostname == master_hostname


def is_clustered():
    from agent.target.cluster import get_cluster_state
    return (get_membership_state() == MEMBERSHIP_STATE_CLUSTERED and
            get_cluster_state() == CLUSTER_STATE_DLM_READY)


def parse_metafile(contents):
    """
    Parse metafile contents to a dict.

    @param      contents - metafile contents as a string

    @return     Metafile contents as a dict.
    """
    d = {}
    for line in contents.splitlines():
        # deal with comments
        ln = line.split("#")[0].strip()
        # parse key/value pairs
        if ln == "":
            continue
        elif "=" in ln and len(ln.split("=")) == 2:
            key, value = map(string.strip, ln.split("="))
            d[key] = value
        else:
            raise Exception("Invalid line in metafile: %s" % line)
    return d


def format_uuid(uuid):
    _uuid = uuid.replace("-", "").lower()
    if len(_uuid) != 32:
        raise RuntimeError('Invalid UUID: %s' % uuid)
    return "-".join([_uuid[0:8], _uuid[8:12], _uuid[12:16],
                     _uuid[16:20], _uuid[20:]])


def write_certificate(filename, certificate):
    certificatefile = open(filename, 'w')
    certificatefile.write(certificate)
    certificatefile.close()


def write_manager_certificate(certificate):
    write_certificate(SSL_MANAGER_CERTIFICATE, certificate)


def write_signed_server_certificate(certificate):
    write_certificate(SSL_SIGNED_SERVER_CERTIFICATE, certificate)


def remove_certificate(certificate):
    try:
        os.unlink(certificate)
    except OSError, err:
        if err.errno != errno.ENOENT:
            raise


def remove_manager_certificate():
    remove_certificate(SSL_MANAGER_CERTIFICATE)


def remove_signed_server_certificate():
    remove_certificate(SSL_SIGNED_SERVER_CERTIFICATE)


def get_certificate(cert_type):
    try:
        certificatefile = open(cert_type)
        certificate = certificatefile.read()
        certificatefile.close()
        return certificate
    except Exception:
        return ''


def get_server_certificate():
    return get_certificate(SSL_SERVER_CERTIFICATE)


def get_signed_server_certificate():
    return get_certificate(SSL_SIGNED_SERVER_CERTIFICATE)


def get_manager_certificate():
    return get_certificate(SSL_MANAGER_CERTIFICATE)
