#!/usr/bin/env python
#
# Copyright (C) 2011-2014 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.
####################################################################################################
#
#	Author:		Joshua Nicholas
#
"""Utilities for parsing a multipath.conf file
"""

import st_utils

import os
import re
import traceback

from blockDeviceOperations	import *

####################################################################################################

__all__ = [	'CCISSdisk_CurState_OBJ'
		,	'CCISSdisk_OBJ'
		,	'CCISS_BlkDev_Operations'
		,	'find_cciss_disks'
		,	'find_particular_cciss_disk' ]

####################################################################################################

class CCISSdisk_CurState_OBJ:

	def __init__(self):
		self.mDisk               = None
		self.mInterface          = None
		self.mBlkDev             = None
		self.mSizeInBytes        = None

	def __repr__(self):
		return "mDisk: %s mInterface: %s mBlkDev: %s mSizeInBytes: %s" % (
						self.mDisk
					,	self.mInterface
					,	self.mBlkDev
					,	self.mSizeInBytes )

class CCISSdisk_OBJ:

	def __init__(self):
		self.mDisk            = None
		self.mCntrlNum        = None
		self.mDiskNum         = None
		self.mInterface       = None
		self.mBlkDev          = None
		self.mDevNums         = None
		self.mDevPath         = None
		self.mWWID            = None
		self.mOldWWID         = None
		self.mBoardID         = None
		self.mFirmwareRev     = None
		self.mModel           = None
		self.mCurState        = None # CCISSdisk_CurState_OBJ

	def __repr__(self):
		return "mDisk: %s mCntrlNum: %s mDiskNum: %s mInterface: %s mBlkDev: %s mDevNums: %s mDevPath: %s mWWID: %s mOldWWID: %s mBoardID: %s mFirmwareRev: %s mModel: %s mCurState: %s" % (
						self.mDisk
					,	self.mCntrlNum
					,	self.mDiskNum
					,	self.mInterface
					,	self.mBlkDev
					,	self.mDevNums
					,	self.mDevPath
					,	self.mWWID
					,	self.mOldWWID
					,	self.mBoardID
					,	self.mFirmwareRev
					,	self.mModel
					,	self.mCurState )

####################################################################################################
class CCISS_BlkDev_Operations(BlkDev_Operations):

	"""CCISS class for BlkDev_Operations
	"""

	########################################
	def __init__( self, blkDevInfo, cciss_disk ):

		if blkDevInfo.mTransport != 'SAS_cciss':
			raise ValueError, "given blkDevInfo %s is not CCISS [%s] !?!" % ( blkDevInfo.mBlkDev, blkDevInfo.mTransport )

		self.mBlkDevInfo	= blkDevInfo
		self.mCCISSdisk		= cciss_disk

	########################################
	def __repr__( self ):

		return "CCISS_BlkDev_Operations[%s]" % ( self.mBlkDevInfo.mBlkDev )

	########################################
	def disconnect_block_device( self, vi_errorAndLog_obj, forceDisconnection ):
		"""
		"""
	#
		vi_errorAndLog_obj.vi_print_wrn( "disconnect_block_device(%s) not implemented for CCISS" % ( self.mBlkDevInfo.mBlkDev ) )
		return 0

	########################################
	def revalidate_block_device( self, vi_errorAndLog_obj ):
		"""
		Updates a standard block device to the latests sysfs values
		"""
	#
		blkdev_info_path = "/sys/block/%s" % ( self.mBlkDevInfo.mBlkdev )
		if not os.path.exists( blkdev_info_path ):
			return 1

		blkdev_info_path = "/sys/block/%s/dev" % ( self.mBlkDevInfo.mBlkdev )
		self.mBlkDevInfo.mDevNums, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
		if read_failed:
			vi_errorAndLog_obj.vi_print_wrn( "blkdev '%s' vanished [%s] !?!" % ( self.mBlkDevInfo.mBlkdev, blkdev_info_path ) )
			return 1

		blkdev_info_path = "/sys/block/%s/size" % ( self.mBlkDevInfo.mBlkdev )
		blkdev_sizeInSectors, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
		if read_failed:
			vi_errorAndLog_obj.vi_print_wrn( "blkdev '%s' vanished [%s] !?!" % ( self.mBlkDevInfo.mBlkdev, blkdev_info_path ) )
			return 1

		self.mBlkDevInfo.mCurState.mSizeInBytes	= int( blkdev_sizeInSectors ) * 512

		return 0

####################################################################################################

global__cached__cciss_disk_DICT		= None

####################################################################################################
def find_cciss_disks( vi_errorAndLog_obj, refresh ):
	"""
	This routine
	"""
#
	global global__cached__cciss_disk_DICT
	if not refresh and global__cached__cciss_disk_DICT:
		return global__cached__cciss_disk_DICT

	if not os.path.exists( '/proc/driver/cciss' ):
		return {}

	cciss_disk_DICT = {}
	try:
		cciss_disk_cmpldRegEx = re.compile( '^cciss\/([a-z]+)(\d+)([a-z]+)(\d+):\s+(\d+\S+)\s+' )
		for cciss_cntlr in os.listdir( '/proc/driver/cciss' ):

			cciss_cntlr_path = "/proc/driver/cciss/%s" % ( cciss_cntlr )
			cciss_cntlr_lines_LIST, read_failed = st_utils.vi_read_text_lines( vi_errorAndLog_obj, cciss_cntlr_path )
			if read_failed:
#				vi_errorAndLog_obj.vi_croak( "Could not extract contents of cciss state file %s" % ( cciss_cntlr_path ) )
				vi_errorAndLog_obj.vi_print_err( "Could not extract contents of cciss state file %s" % ( cciss_cntlr_path ) )

			model_cmpldRegEx = re.compile( "^%s: " % ( cciss_cntlr ) )
			cciss_model      = None
			for cciss_cntlr_line in cciss_cntlr_lines_LIST:
				if model_cmpldRegEx.match( cciss_cntlr_line ):
					cciss_model = model_cmpldRegEx.sub( '', cciss_cntlr_line )
					break
			if not cciss_model:
				vi_errorAndLog_obj.vi_print_wrn( "Unexpected cciss state file format %s" % ( cciss_cntlr_path ) )
				cciss_model = 'Unknown'

			board_id_cmpldRegEx = re.compile( '^Board ID: ' )
			cciss_board_id = None
			for cciss_cntlr_line in cciss_cntlr_lines_LIST:
				if board_id_cmpldRegEx.match( cciss_cntlr_line ):
					cciss_board_id = board_id_cmpldRegEx.sub( '', cciss_cntlr_line )
					break
			if not cciss_board_id:
				vi_errorAndLog_obj.vi_print_wrn( "Unexpected cciss state file format %s" % ( cciss_cntlr_path ) )
				cciss_board_id = 'Unknown'

			firmware_ver_cmpldRegEx = re.compile( '^Firmware Version: ' )
			cciss_firmware_ver = None
			for cciss_cntlr_line in cciss_cntlr_lines_LIST:
				if firmware_ver_cmpldRegEx.match( cciss_cntlr_line ):
					cciss_firmware_ver = firmware_ver_cmpldRegEx.sub( '', cciss_cntlr_line )
					break
			if not cciss_firmware_ver:
				vi_errorAndLog_obj.vi_print_wrn( "Unexpected cciss state file format %s" % ( cciss_cntlr_path ) )
				cciss_firmware_ver = 'Unknown'

			for cciss_cntlr_line in cciss_cntlr_lines_LIST:

				cciss_disk_match = cciss_disk_cmpldRegEx.match( cciss_cntlr_line )
				if not cciss_disk_match:
					continue

				cciss_disk_name      = "%s%s%s%s" % ( cciss_disk_match.group(1), cciss_disk_match.group(2), cciss_disk_match.group(3), cciss_disk_match.group(4) )
				cciss_disk_cntlr_num = cciss_disk_match.group(2)
				cciss_disk_disk_num  = cciss_disk_match.group(4)
				cciss_blkdev         = "cciss!%s" % ( cciss_disk_name )

				blkdev_info_path = "/sys/block/%s/dev" % ( cciss_blkdev )
				cciss_disk_devNums, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
				if read_failed:
					vi_errorAndLog_obj.vi_print_wrn( "blkdev '%s' vanished [%s] !?!" % ( cciss_blkdev, blkdev_info_path ) )
					continue

				blkdev_info_path = "/sys/block/%s/size" % ( cciss_blkdev )
				cciss_disk_sizeInSectors, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
				if read_failed:
					vi_errorAndLog_obj.vi_print_wrn( "blkdev '%s' vanished [%s] !?!" % ( cciss_blkdev, blkdev_info_path ) )
					continue
				cciss_disk_sizeInBytes = int(cciss_disk_sizeInSectors) * 512

				cmd_LIST	= [ 'scsi_id', '-n', '-g', '-u', '-d', "/dev/%s" % ( cciss_blkdev ) ]

				stdout_LIST, stderr_LIST, exitStatus, deltaTime = st_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST )

				cciss_disk_wwid = stdout_LIST[ 0 ].strip()

				stdout_LIST, stderr_LIST, exitStatus, deltaTime = st_utils.vi_exe_sync( vi_errorAndLog_obj, [ 'vi_compute_old_cciss_id.sh', cciss_blkdev ] )

				cciss_disk_old_wwid = stdout_LIST[ 0 ].strip()

				cciss_disk_curState                 = CCISSdisk_CurState_OBJ()
				cciss_disk_curState.mDisk           = cciss_disk_name
				cciss_disk_curState.mInterface      = cciss_cntlr
				cciss_disk_curState.mBlkDev         = cciss_blkdev
				cciss_disk_curState.mSizeInBytes    = cciss_disk_sizeInBytes

				cciss_disk                  = CCISSdisk_OBJ()
				cciss_disk.mDisk            = cciss_disk_name
				cciss_disk.mCntrlNum        = cciss_disk_cntlr_num
				cciss_disk.mDiskNum         = cciss_disk_disk_num
				cciss_disk.mInterface       = cciss_cntlr
				cciss_disk.mBlkDev          = cciss_blkdev
				cciss_disk.mDevNums         = cciss_disk_devNums
				cciss_disk.mDevPath         = "/dev/cciss/%s" % ( cciss_disk_name )
				cciss_disk.mWWID            = cciss_disk_wwid
				cciss_disk.mOldWWID         = cciss_disk_old_wwid
				cciss_disk.mBoardID         = cciss_board_id
				cciss_disk.mFirmwareRev     = cciss_firmware_ver
				cciss_disk.mModel           = cciss_model
				cciss_disk.mCurState        = cciss_disk_curState

				cciss_disk_DICT[ cciss_blkdev ]	= cciss_disk
#				vi_errorAndLog_obj.vi_print_dbg( 0, cciss_disk )

	except:
#		vi_errorAndLog_obj.vi_croak_exception( traceback.format_exc() )
		vi_errorAndLog_obj.vi_print_err( "Problems extracting cciss disk info\n%s" % ( traceback.format_exc() ) )
		return []

	if 0:
		vi_errorAndLog_obj.vi_print_dbg( 0, cciss_disk_DICT )

	global__cached__cciss_disk_DICT = cciss_disk_DICT
	return global__cached__cciss_disk_DICT

####################################################################################################
def find_particular_cciss_disk( vi_errorAndLog_obj, cciss_blkDev, refresh ):
	"""
	This routine
	"""
#
	cciss_disk_DICT	= find_cciss_disks( vi_errorAndLog_obj, refresh )

	return cciss_disk_DICT.get( cciss_blkDev, None )

####################################################################################################
# END
