#!/usr/bin/env python
#
# Copyright (C) 2011-2014 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.
####################################################################################################
#
#	Author:		Joshua Nicholas
#
"""Utilities for dealing with the multipath
"""

import st_utils

import re
import pickle
import itertools
import traceback
from itertools import chain

####################################################################################################

__all__ = [	'global_multipathConf_default_callout'
		,	'multipathBlkDevInfo_OBJ'
		,	'multipathConfDeviceSpec'
		,	'multipathConfBlacklistDevice'
		,	'get_multipath_info'
		,	'resize_multipath_maps'
		,	'ensure_multipath_maps'
		,	'parse_multipath_conf_file'
		,	'get_multipath_conf' ]

####################################################################################################

class multipathBlkDevInfo_OBJ:

	def __init__(self):
		self.mWWID              = None
		self.mDevMapNum         = None
		self.mSCSIdevIDstr      = None
		self.mBlkDev            = None
		self.mDevNums           = None
		self.mPathDMmapState    = None
		self.mCheckerState      = None
		self.mSizeInSect		= -1

	def __repr__(self):
		return "mWWID: %s mDevMapNum: %s mSCSIdevIDstr: %s mBlkDev: %s mDevNums: %s mPathDMmapState: %s mCheckerState: %s mSizeInSect:%d" % (
						self.mWWID
					,	self.mDevMapNum
					,	self.mSCSIdevIDstr
					,	self.mBlkDev
					,	self.mDevNums
					,	self.mPathDMmapState
					,	self.mCheckerState
					,	self.mSizeInSect )

class multipathConfDeviceSpec:

	def __init__( self ):
		self.mIsExplicit			= False
		self.mIsCalloutGenerated	= False
		self.mVendor				= None
		self.mVendor_cmpldRegEx		= None
		self.mProduct				= None
		self.mProduct_cmpldRegEx	= None
		self.mCallout				= None
		self.mBlacklist				= None

	def __repr__( self ):
		if self.mIsExplicit == False:
			return "-v: %s p: %s c: %s b: %s" % (
							self.mVendor
						,	self.mProduct
						,	self.mCallout
						,	self.mBlacklist )
		else:
			return "+v: %s p: %s c: %s b: %s" % (
							self.mVendor
						,	self.mProduct
						,	self.mCallout
						,	self.mBlacklist )

	def __cmp__( self, other ):
		if not isinstance( other, multipathConfDeviceSpec ):
			return -1

		cmpVal	= cmp( self.mVendor, other.mVendor )
		if cmpVal != 0:
			return cmpVal

		cmpVal	= cmp( self.mProduct, other.mProduct )
		if cmpVal != 0:
			return cmpVal

		cmpVal	= cmp( self.mCallout, other.mCallout )
		if cmpVal != 0:
			return cmpVal

		cmpVal	= cmp( self.mBlacklist, other.mBlacklist )
		if cmpVal != 0:
			return cmpVal

		return 0

class multipathConfBlacklistDevNode:

	def __init__( self, theDevNode ):
		self.mDevNode				= theDevNode
		self.mDevNode_cmpldRegEx	= None

	def __repr__( self ):
		return "%s" % ( self.mDevNode )

class multipathConfBlacklistDevice:

	def __init__( self ):
		self.mVendor				= None
		self.mVendor_cmpldRegEx		= None
		self.mProduct				= None
		self.mProduct_cmpldRegEx	= None

	def __repr__( self ):
		return "v: %s p: %s" % (
						self.mVendor
					,	self.mProduct )

	def __cmp__(self, other):
		if not isinstance( other, multipathConfBlacklistDevice ):
			return -1
		if self.mVendor < other.mVendor:
			return -1
		if self.mVendor > other.mVendor:
			return  1
		if self.mProduct < other.mProduct:
			return -1
		if self.mProduct > other.mProduct:
			return  1
		return 0

class multipathConfBlacklistWWID:

	def __init__( self, theWWID ):
		self.mWWID				= theWWID
		self.mWWID_cmpldRegEx	= None

	def __repr__( self ):
		return "%s" % ( self.mWWID )

class multipathConfBlacklistInfo:

	def __init__( self ):
		self.mDevNode_LIST	= []
		self.mDevice_LIST	= []
		self.mWWID_LIST		= []

	def __repr__( self ):
		return "devNode_L: %s device_L: %s wwid_L: %s" % (
						self.mDevNode_LIST
					,	self.mDevice_LIST
					,	self.mWWID_LIST )

####################################################################################################

global_multipathConf_default_callout  	= 'scsi_id -g -u -d /dev/%n'

global_multipath_merged_conf_cache_path	= '/var/cache/ovs-multipath/multipath_merged_conf_cache.pkl'
global_multipath_etc_conf_cache_path	= '/var/cache/ovs-multipath/multipath_etc_conf_cache.pkl'

global__cached__multipath_info_DICT		= None

####################################################################################################
def get_block_device_size ( vi_errorAndLog_obj, name ):
	"""
	This routine returns the size of the block device in sector size.
	"""

	if not name:
		return -1

	sysfs_attr = "/sys/block/%s/size" % (name)

	str_size, read_failed = st_utils.vi_read_text_value(vi_errorAndLog_obj, sysfs_attr)

	if read_failed:
		return -1

	return int(str_size)

####################################################################################################
def get_multipath_info( vi_errorAndLog_obj, refresh ):
	"""
	This routine
	"""
#
	global global__cached__multipath_info_DICT
	if not refresh and global__cached__multipath_info_DICT:
		return global__cached__multipath_info_DICT

	multipathInfo_DICT = {}
	# 3600805f3001321c02f8962adb414003c dm-11 COMPAQ,MSA1000 VOLUME
	mp_wwid_line_cmpldRegEx  = re.compile( '^(.*)\sdm-(\d+)\s+' )
	# Old line:  \_ 9:0:0:2  sdc 8:32  [active][ready]
	mp_path_line_cmpldRegEx_old  = re.compile( '^\\\\_\s+(\d+:\d+:\d+:\d+)\s+(\S*)\s+(\d+:\d+)\s+\[(\w+)\]\[(\w+)\]$' )
	# New line: `- 3:0:0:2  sdb 8:16  active ready running
	mp_path_line_cmpldRegEx_new  = re.compile( '^\`-\s+(\d+:\d+:\d+:\d+)\s+(\S*)\s+(\d+:\d+)\s+(\w+)\s+(\w+)\s+(\w+)$' )


	mp_path_line_cmpldRegEx_1 = re.compile( '^\`-\s+(\d+:\d+:\d+:\d+)\s+(\S*)\s+(\d+:\d+)\s+(\w+)\s+(\w+)\s+(\w+)$' )
	mp_path_line_cmpldRegEx_2 = re.compile( '^\|-\s+(\d+:\d+:\d+:\d+)\s+(\S*)\s+(\d+:\d+)\s+(\w+)\s+(\w+)\s+(\w+)$' )
	mp_path_line_cmpldRegEx_3 = re.compile( '\| `-\s+(\d+:\d+:\d+:\d+)\s+(\S*)\s+(\d+:\d+)\s+(\w+)\s+(\w+)\s+(\w+)$' )
	mp_path_line_cmpldRegEx_4 = re.compile( '\| \|-\s+(\d+:\d+:\d+:\d+)\s+(\S*)\s+(\d+:\d+)\s+(\w+)\s+(\w+)\s+(\w+)$' )

	stdout_LIST, stderr_LIST, exitStatus, deltaTime = st_utils.vi_exe_sync( vi_errorAndLog_obj, 'multipathd list topology' )
	del( stderr_LIST )

	multipath_lines_LIST = map( str.strip, stdout_LIST )
	del( stdout_LIST )
	try:
		cur_wwid = ''
		cur_dev_map = -1
		multipath_line = ''
		for multipath_line in multipath_lines_LIST:
			if len(multipath_line) < 1:
				continue

			#vi_errorAndLog_obj.vi_print_dbg( 0, "multipath_line: %s" % ( multipath_line ) )

			if multipath_line.startswith("size"):
				continue
		        if multipath_line.startswith("create:"):
                                multipath_line = multipath_line.lstrip("create:").strip()

			# Old format had '\' in front of block devices. New has the `
			if multipath_line[0] != '|' and multipath_line[0] != '`':
				mp_wwid_line_match = mp_wwid_line_cmpldRegEx.match( multipath_line )
				if not mp_wwid_line_match:
					continue
				cur_wwid = mp_wwid_line_match.group( 1 )
				cur_dev_map_num = int( mp_wwid_line_match.group( 2 ) )
				#vi_errorAndLog_obj.vi_print_dbg( 0,"WWID: %s    dm-%s" % ( cur_wwid, cur_dev_map_num ) )
                            
				continue
			mp_path_line_match = mp_path_line_cmpldRegEx_1.match( multipath_line )
			if not mp_path_line_match:
				mp_path_line_match = mp_path_line_cmpldRegEx_2.match(multipath_line)
				if not mp_path_line_match:
					mp_path_line_match = mp_path_line_cmpldRegEx_3.match(multipath_line)
					if not mp_path_line_match:
						mp_path_line_match = mp_path_line_cmpldRegEx_4.match(multipath_line)
						if not mp_path_line_match:
							continue

			mpBlkDevInfo                    = multipathBlkDevInfo_OBJ()
			mpBlkDevInfo.mWWID              = cur_wwid
			mpBlkDevInfo.mDevMapNum         = cur_dev_map_num
			mpBlkDevInfo.mSCSIdevIDstr      = mp_path_line_match.group( 1 )
			mpBlkDevInfo.mBlkDev            = mp_path_line_match.group( 2 )
			mpBlkDevInfo.mDevNums           = mp_path_line_match.group( 3 )
			mpBlkDevInfo.mPathDMmapState    = mp_path_line_match.group( 4 )
			mpBlkDevInfo.mCheckerState      = mp_path_line_match.group( 5 )
			mpBlkDevInfo.mSizeInSect		= get_block_device_size(vi_errorAndLog_obj, mpBlkDevInfo.mBlkDev)
			#vi_errorAndLog_obj.vi_print_dbg( 0, mpBlkDevInfo )

			blkDev_DICT = multipathInfo_DICT.get( cur_wwid, None )
			if not blkDev_DICT:
				multipathInfo_DICT[ cur_wwid ] = { mpBlkDevInfo.mBlkDev : mpBlkDevInfo }
#				dev_map_num_to_wwid_DICT[ cur_dev_map_num ] = cur_wwid
#				vi_errorAndLog_obj.vi_print_dbg( 0, "ADDED:  '%s' dm-%s '%s'" % ( cur_wwid, cur_dev_map_num, mpBlkDevInfo.mBlkDev ) )
			else:
				blkDev_DICT[ mpBlkDevInfo.mBlkDev ] = mpBlkDevInfo
#				vi_errorAndLog_obj.vi_print_dbg( 0, "ADDED:  '%s' '%s'" % ( cur_wwid, mpBlkDevInfo.mBlkDev ) )

		del( multipath_lines_LIST )
	except:
#		vi_errorAndLog_obj.vi_croak_exception( traceback.format_exc() )
		vi_errorAndLog_obj.vi_print_err( "multipathd list topology format %s '%s'\n%s" % ( cur_wwid, multipath_line, traceback.format_exc() ) )
		del( multipath_lines_LIST )
		return {}

	if 0:
		sorted_wwid_LIST = multipathInfo_DICT.keys()
		sorted_wwid_LIST.sort()
		for wwid in sorted_wwid_LIST:
			wwid_DICT = multipathInfo_DICT[ wwid ]
			vi_errorAndLog_obj.vi_print_dbg( 0, wwid, wwid_DICT )

	global__cached__multipath_info_DICT = multipathInfo_DICT
	return global__cached__multipath_info_DICT

####################################################################################################
def parse_multipath_conf_defaults( vi_errorAndLog_obj, mpc_token_itr, mpc_token_cntr ):
	"""
	This routine
	"""
#
	mpConf_default_callout = global_multipathConf_default_callout

	while True:
		try:
			mpc_key_token = mpc_token_itr.next()
			mpc_token_cntr.next()
			if  '}' == mpc_key_token:
				return mpConf_default_callout, 0
			mpc_val_token = mpc_token_itr.next()
			mpc_token_cntr.next()
		except StopIteration:
#			vi_errorAndLog_obj.vi_croak( "multipathCnfLines contains unexpected early end [%s]" % ( mpc_token_cntr ) )
			vi_errorAndLog_obj.vi_print_err( "multipathCnfLines contains unexpected early end [%s]" % ( mpc_token_cntr ) )
			return mpConf_default_callout, 1

#		vi_errorAndLog_obj.vi_print_dbg( 0, "%s ==> %s" % ( mpc_key_token, mpc_val_token ) )

		if 'getuid_callout' == mpc_key_token:
			mpConf_default_callout = mpc_val_token

####################################################################################################
def parse_multipath_conf_blacklist_device( vi_errorAndLog_obj, mpc_token_itr, mpc_token_cntr ):
	"""
	This routine
	"""
#
	mpConf_blkLstDev = multipathConfBlacklistDevice()

	while True:
		try:
			mpc_key_token = mpc_token_itr.next()
			mpc_token_cntr.next()
			if  '}' == mpc_key_token:
				break
			mpc_val_token = mpc_token_itr.next()
			mpc_token_cntr.next()
		except StopIteration:
#			vi_errorAndLog_obj.vi_croak( "multipathCnfLines contains unexpected early end [%s]" % ( mpc_token_cntr ) )
			vi_errorAndLog_obj.vi_print_err( "multipathCnfLines contains unexpected early end [%s]" % ( mpc_token_cntr ) )
			return None, 1

#		vi_errorAndLog_obj.vi_print_dbg( 0, "%s ==> %s" % ( mpc_key_token, mpc_val_token ) )

		if  'vendor' == mpc_key_token:
			mpConf_blkLstDev.mVendor = mpc_val_token

		if  'product' == mpc_key_token:
			mpConf_blkLstDev.mProduct = mpc_val_token

	if not mpConf_blkLstDev.mVendor:
#		vi_errorAndLog_obj.vi_croak( "multipathCnfLines incomplete blacklist device :: vendor [%s]" % ( mpc_token_cntr ) )
		vi_errorAndLog_obj.vi_print_err( "multipathCnfLines incomplete blacklist device :: vendor [%s]" % ( mpc_token_cntr ) )
		return None, 1

	if not mpConf_blkLstDev.mProduct:
		mpConf_blkLstDev.mProduct = '*'

	return mpConf_blkLstDev, 0

####################################################################################################
def parse_multipath_conf_blacklist( vi_errorAndLog_obj, mpc_token_itr, mpc_token_cntr ):
	"""
	This routine
	"""
#
	numOfParsingErrs		= 0
	mpConf_blacklistInfo	= multipathConfBlacklistInfo()
	while True:
		try:
			mpc_key_token = mpc_token_itr.next()
			mpc_token_cntr.next()
			if  '}' == mpc_key_token:
				return mpConf_blacklistInfo, numOfParsingErrs
			mpc_val_token = mpc_token_itr.next()
			mpc_token_cntr.next()
		except StopIteration:
#			vi_errorAndLog_obj.vi_croak( "multipathCnfLines contains unexpected early end [%s]" % ( mpc_token_cntr ) )
			vi_errorAndLog_obj.vi_print_err( "multipathCnfLines contains unexpected early end [%s]" % ( mpc_token_cntr ) )
			numOfParsingErrs	= numOfParsingErrs + 1
			return mpConf_blacklistInfo, numOfParsingErrs

#		vi_errorAndLog_obj.vi_print_dbg( 0, "%s ==> %s" % ( mpc_key_token, mpc_val_token ) )

		if 'devnode' == mpc_key_token:
			if mpc_val_token[0] == '*':
				mpc_val_token = '.' + mpc_val_token
			mpConf_blacklistInfo.mDevNode_LIST.append( multipathConfBlacklistDevNode( mpc_val_token ) )
			continue

		if 'device' == mpc_key_token:
			mpConf_blkLstDev, additionalParsingErrs	= parse_multipath_conf_blacklist_device( vi_errorAndLog_obj, mpc_token_itr, mpc_token_cntr )
			numOfParsingErrs						= numOfParsingErrs + additionalParsingErrs
			if mpConf_blkLstDev:
				if not mpConf_blkLstDev in mpConf_blacklistInfo.mDevice_LIST:
					mpConf_blacklistInfo.mDevice_LIST.append( mpConf_blkLstDev )
			continue

		if 'wwid' == mpc_key_token:
			mpConf_blacklistInfo.mWWID_LIST.append( multipathConfBlacklistWWID( mpc_val_token ) )
			continue

                if '%include' == mpc_key_token:
                        if mpc_val_token:
                                mpc_include_LIST, read_failed = st_utils.vi_read_text_lines( vi_errorAndLog_obj, mpc_val_token )
                                if read_failed:
                                        continue
                                mpc_include_itr = iter(mpc_include_LIST) 
                                mpc_token_itr = chain(mpc_token_itr, mpc_include_itr)
                        continue

#		vi_errorAndLog_obj.vi_croak( "multipathCnfLines contains unexpected format '%s' => '%s' [%s]" % ( mpc_key_token, mpc_val_token, mpc_token_cntr ) )
		vi_errorAndLog_obj.vi_print_err( "multipathCnfLines contains unexpected format '%s' => '%s' [%s]" % ( mpc_key_token, mpc_val_token, mpc_token_cntr ) )
		numOfParsingErrs	= numOfParsingErrs + 1
		return mpConf_blacklistInfo, numOfParsingErrs

####################################################################################################
def parse_multipath_conf_blacklist_exceptions( vi_errorAndLog_obj, mpc_token_itr, mpc_token_cntr ):
	"""
	This routine
	"""
#
	return parse_multipath_conf_blacklist( vi_errorAndLog_obj, mpc_token_itr, mpc_token_cntr )

####################################################################################################
def parse_multipath_conf_device( vi_errorAndLog_obj, mpc_token_itr, mpc_token_cntr ):
	"""
	This routine
	"""
#
	mpConf_devSpec = multipathConfDeviceSpec()

	while True:
		try:
			mpc_key_token = mpc_token_itr.next()
			mpc_token_cntr.next()
			if  '}' == mpc_key_token:
				break
			mpc_val_token = mpc_token_itr.next()
			mpc_token_cntr.next()
		except StopIteration:
#			vi_errorAndLog_obj.vi_croak( "multipathCnfLines contains unexpected early end [%s]" % ( mpc_token_cntr ) )
			vi_errorAndLog_obj.vi_print_err( "multipathCnfLines contains unexpected early end [%s]" % ( mpc_token_cntr ) )
			return None, 1

#		vi_errorAndLog_obj.vi_print_dbg( 0, "%s ==> %s" % ( mpc_key_token, mpc_val_token ) )

		if  'vendor' == mpc_key_token:
			mpConf_devSpec.mVendor = mpc_val_token

		if  'product' == mpc_key_token:
			mpConf_devSpec.mProduct = mpc_val_token

		if  'getuid_callout' == mpc_key_token:
			mpConf_devSpec.mCallout = mpc_val_token
			if mpc_val_token.find( ' --force_uniq ' ) >= 0:
				mpConf_devSpec.mIsCalloutGenerated	= True

		if  'product_blacklist' == mpc_key_token:
			mpConf_devSpec.mBlacklist = mpc_val_token

	if not mpConf_devSpec.mVendor:
#		vi_errorAndLog_obj.vi_croak( "multipathCnfLines incomplete device spec :: vendor [%s]" % ( mpc_token_cntr ) )
		vi_errorAndLog_obj.vi_print_err( "multipathCnfLines incomplete device spec :: vendor [%s]" % ( mpc_token_cntr ) )
		return None, 1

	if not mpConf_devSpec.mProduct:
		mpConf_devSpec.mProduct = '*'

	return mpConf_devSpec, 0

####################################################################################################
def parse_multipath_conf_devices( vi_errorAndLog_obj, mpc_token_itr, mpc_token_cntr ):
	"""
	This routine
	"""
#
	numOfParsingErrs	= 0
	mpConf_devSpec_LIST	= []

	while True:
		try:
			mpc_token = mpc_token_itr.next()
			mpc_token_cntr.next()
			if  '}' == mpc_token:
				return mpConf_devSpec_LIST, numOfParsingErrs

			if 'device' != mpc_token:
#				vi_errorAndLog_obj.vi_croak( "multipath.conf contains unexpected format '%s' [%s]" % ( mpc_token, mpc_token_cntr ) )
				vi_errorAndLog_obj.vi_print_err( "multipath.conf contains unexpected format '%s' [%s]" % ( mpc_token, mpc_token_cntr ) )
				numOfParsingErrs	= numOfParsingErrs + 1
				return mpConf_devSpec_LIST, numOfParsingErrs

			mpc_token = mpc_token_itr.next()
			mpc_token_cntr.next()
			if '{' != mpc_token:
#				vi_errorAndLog_obj.vi_croak( "multipath.conf contains unexpected format '%s' [%s]" % ( mpc_token, mpc_token_cntr ) )
				vi_errorAndLog_obj.vi_print_err( "multipath.conf contains unexpected format '%s' [%s]" % ( mpc_token, mpc_token_cntr ) )
				numOfParsingErrs	= numOfParsingErrs + 1
				return mpConf_devSpec_LIST, numOfParsingErrs

		except StopIteration:
#			vi_errorAndLog_obj.vi_croak( "multipath.conf contains unexpected early end [%s]" % mpc_token_cntr )
			vi_errorAndLog_obj.vi_print_err( "multipath.conf contains unexpected early end [%s]" % mpc_token_cntr )
			numOfParsingErrs	= numOfParsingErrs + 1
			return mpConf_devSpec_LIST, numOfParsingErrs

		mpConf_devSpec, additionalParsingErrs	= parse_multipath_conf_device( vi_errorAndLog_obj, mpc_token_itr, mpc_token_cntr )
		numOfParsingErrs						= numOfParsingErrs + additionalParsingErrs
		if not mpConf_devSpec:
			continue

#		vi_errorAndLog_obj.vi_print_dbg( 0, mpConf_devSpec )

		mpConf_devSpec_LIST.append( mpConf_devSpec )

	return mpConf_devSpec_LIST, numOfParsingErrs

####################################################################################################
def parse_multipath_conf_multipaths( vi_errorAndLog_obj, mpc_token_itr, mpc_token_cntr ):
	"""
	This routine
	"""
#
	while True:
		try:
			mpc_key_token = mpc_token_itr.next()
			mpc_token_cntr.next()
			if  '}' == mpc_key_token:
				return 0
			mpc_val_token = mpc_token_itr.next()
			mpc_token_cntr.next()
		except StopIteration:
#			vi_errorAndLog_obj.vi_croak( "multipath.conf contains unexpected early end [%s]" % ( mpc_token_cntr ) )
			vi_errorAndLog_obj.vi_print_err( "multipath.conf contains unexpected early end [%s]" % ( mpc_token_cntr ) )
			return 1

		vi_errorAndLog_obj.vi_print_err( "found unsupported multipath '%s' '%s' %s" % ( mpc_key_token, mpc_val_token, mpc_token_cntr ) )
		return 1

####################################################################################################
def parse_multipath_conf_lines_list( vi_errorAndLog_obj, mpConfLines_LIST ):
	"""
	This routine
	"""
#
	# Remove comments, empty lines
	# Tokenize based on spaces and "quoted" strings
	#     AND if not quoted all tokens after the first in a line are considered part of the same token
	mpc_tokens	= []
	cur_token	= ''
	is_quoted	= False
	for line in mpConfLines_LIST:
		conf_line = (line.split('#',1))[0].lstrip().rstrip()
		if not conf_line:
			continue
		is_first_token_in_line	= True
		was_quoted				= False
		for mpc_char in conf_line:
			if mpc_char == '"':
				if is_quoted:
					is_quoted	= False
					was_quoted	= True
				else:
					is_quoted	= True
				continue
			if not mpc_char.isspace() or is_quoted:
				cur_token += mpc_char
				continue
			if len( cur_token ) < 1:
				continue
			if not is_first_token_in_line and not was_quoted:
				cur_token += mpc_char
				continue
#			vi_errorAndLog_obj.vi_print_dbg( 0, ">%s<" % ( cur_token ), None )
			mpc_tokens.append( cur_token )
			cur_token = ''
			is_first_token_in_line	= False
		if len( cur_token ) > 0 :
#			vi_errorAndLog_obj.vi_print_dbg( 0, "<%s>" % ( cur_token ), None )
			mpc_tokens.append( cur_token )
			cur_token = ''
#		vi_errorAndLog_obj.vi_print_dbg( 0, '' )

#	vi_errorAndLog_obj.vi_print_dbg( 0, mpc_tokens )
#	vi_errorAndLog_obj.vi_stop( 0 )

	numOfParsingErrs				= 0
	mpConf_default_callout			= global_multipathConf_default_callout
	mpConf_blacklistInfo			= multipathConfBlacklistInfo()
	mpConf_blacklistExceptionsInfo	= multipathConfBlacklistInfo()
	mpConf_devSpec_LIST				= []

	mpc_token_itr  = iter(mpc_tokens)
	mpc_token_cntr = itertools.count(0)
	while True:
		try:
			mpc_tok_start = mpc_token_itr.next()
			mpc_token_cntr.next()
			if not re.match( '^\w+$', mpc_tok_start ):
#				vi_errorAndLog_obj.vi_croak( 'multipathCnfLines contains unexpected format 1 [%s]' % ( mpc_token_cntr ) )
				vi_errorAndLog_obj.vi_print_err( "multipathCnfLines contains unexpected format 1 [%s]" % ( mpc_token_cntr ) )
				numOfParsingErrs	= numOfParsingErrs + 1
				continue
		except StopIteration:
			break

		try:
			mpc_token = mpc_token_itr.next()
			mpc_token_cntr.next()
			if '{' != mpc_token:
#				vi_errorAndLog_obj.vi_croak( 'multipathCnfLines contains unexpected format 2 [%s]' % ( mpc_token_cntr ) )
				vi_errorAndLog_obj.vi_print_err( "multipathCnfLines contains unexpected format 2 [%s]" % ( mpc_token_cntr ) )
				numOfParsingErrs	= numOfParsingErrs + 1
				continue
		except StopIteration:
#			vi_errorAndLog_obj.vi_croak( 'multipathCnfLines contains unexpected early end [%s]' % ( mpc_token_cntr ) )
			vi_errorAndLog_obj.vi_print_err( "multipathCnfLines contains unexpected early end [%s]" % ( mpc_token_cntr ) )
			numOfParsingErrs	= numOfParsingErrs + 1
			break

#		vi_errorAndLog_obj.vi_print_dbg( 0, "BEG Tok '%s' %s" % ( mpc_tok_start, mpc_token_cntr ) )

		if 'defaults' == mpc_tok_start:
			mpConf_default_callout, additionalParsingErrs	= parse_multipath_conf_defaults( vi_errorAndLog_obj, mpc_token_itr, mpc_token_cntr )
			numOfParsingErrs								= numOfParsingErrs + additionalParsingErrs
#			vi_errorAndLog_obj.vi_print_dbg( 0, "END Tok '%s' %s" % ( mpc_tok_start, mpc_token_cntr ) )
			continue

		if 'blacklist' == mpc_tok_start:
			mpConf_blacklistInfo, additionalParsingErrs	= parse_multipath_conf_blacklist( vi_errorAndLog_obj, mpc_token_itr, mpc_token_cntr )
			numOfParsingErrs							= numOfParsingErrs + additionalParsingErrs
#			vi_errorAndLog_obj.vi_print_dbg( 0, "END Tok '%s' %s" % ( mpc_tok_start, mpc_token_cntr ) )
			continue

		if 'blacklist_exceptions' == mpc_tok_start:
			mpConf_blacklistExceptionsInfo, additionalParsingErrs	= parse_multipath_conf_blacklist_exceptions( vi_errorAndLog_obj, mpc_token_itr, mpc_token_cntr )
			numOfParsingErrs										= numOfParsingErrs + additionalParsingErrs
#			vi_errorAndLog_obj.vi_print_dbg( 0, "END Tok '%s' %s" % ( mpc_tok_start, mpc_token_cntr ) )
			continue

		if 'devices' == mpc_tok_start:
			mpConf_devSpec_LIST, additionalParsingErrs	= parse_multipath_conf_devices( vi_errorAndLog_obj, mpc_token_itr, mpc_token_cntr )
			numOfParsingErrs							= numOfParsingErrs + additionalParsingErrs
#			vi_errorAndLog_obj.vi_print_dbg( 0, "END Tok '%s' %s" % ( mpc_tok_start, mpc_token_cntr ) )
			continue

		if 'multipaths' == mpc_tok_start:
			additionalParsingErrs	= parse_multipath_conf_multipaths( vi_errorAndLog_obj, mpc_token_itr, mpc_token_cntr )
			numOfParsingErrs		= numOfParsingErrs + additionalParsingErrs
#			vi_errorAndLog_obj.vi_print_dbg( 0, "END Tok '%s' %s" % ( mpc_tok_start, mpc_token_cntr ) )
			continue

#		vi_errorAndLog_obj.vi_croak( "multipathCnfLines contains unexpected section '%s' [%s]" % ( mpc_tok_start, mpc_token_cntr ) )
		vi_errorAndLog_obj.vi_print_err( "multipathCnfLines contains unexpected section '%s' [%s]" % ( mpc_tok_start, mpc_token_cntr ) )
		numOfParsingErrs	= numOfParsingErrs + 1

	for mpConf_blkLstDevNode in mpConf_blacklistInfo.mDevNode_LIST:
		mpConf_blkLstDevNode.mDevNode_cmpldRegEx	= re.compile( mpConf_blkLstDevNode.mDevNode )

	for mpConf_blkLstDev in mpConf_blacklistInfo.mDevice_LIST:
		try:
			mpConf_blkLstDev.mVendor_cmpldRegEx			= re.compile( mpConf_blkLstDev.mVendor )
		except:
			mpConf_blkLstDev.mVendor_cmpldRegEx            = re.compile( '.*?' )
		try:
			mpConf_blkLstDev.mProduct_cmpldRegEx		= re.compile( mpConf_blkLstDev.mProduct )
		except:
 			mpConf_blkLstDev.mProduct_cmpldRegEx           = re.compile( '.*?' )

	for mpConf_blkLstWWID in mpConf_blacklistInfo.mWWID_LIST:
		try:
			mpConf_blkLstWWID.mWWID_cmpldRegEx			= re.compile( mpConf_blkLstWWID.mWWID )
		except:
			pass

	for mpConf_blkLstDevNode in mpConf_blacklistExceptionsInfo.mDevNode_LIST:
		try:
			mpConf_blkLstDevNode.mDevNode_cmpldRegEx	= re.compile( mpConf_blkLstDevNode.mDevNode )
		except:
			pass

	for mpConf_blkLstDev in mpConf_blacklistExceptionsInfo.mDevice_LIST:
		try:
			mpConf_blkLstDev.mVendor_cmpldRegEx			= re.compile( mpConf_blkLstDev.mVendor )
			mpConf_blkLstDev.mProduct_cmpldRegEx		= re.compile( mpConf_blkLstDev.mProduct )
		except:
			pass

	for mpConf_blkLstWWID in mpConf_blacklistExceptionsInfo.mWWID_LIST:
		try:
			mpConf_blkLstWWID.mWWID_cmpldRegEx			= re.compile( mpConf_blkLstWWID.mWWID )
		except:
			pass

	for mpConf_devSpec in mpConf_devSpec_LIST:
		try:
			mpConf_devSpec.mVendor_cmpldRegEx			= re.compile( mpConf_devSpec.mVendor )
			mpConf_devSpec.mProduct_cmpldRegEx			= re.compile( mpConf_devSpec.mProduct )
		except:
			pass

	return numOfParsingErrs, mpConf_default_callout, mpConf_blacklistInfo, mpConf_blacklistExceptionsInfo, mpConf_devSpec_LIST

####################################################################################################
def parse_multipath_conf_file( vi_errorAndLog_obj, multipathConfFilePath ):
	"""
	This routine
	"""
#
	if multipathConfFilePath == None:
		multipathConfFilePath	= '/etc/multipath.conf'

	if multipathConfFilePath == '/etc/multipath.conf':
		multipath_etc_conf_cache_FILE		= None
		try:
			multipath_etc_conf_cache_FILE	= open( global_multipath_etc_conf_cache_path, 'rb' )
			mpConf_default_callout, mpConf_blacklistInfo, mpConf_blacklistExceptionsInfo, mpConf_devSpec_LIST = pickle.load( multipath_etc_conf_cache_FILE )
			multipath_etc_conf_cache_FILE.close()
		except:
			if multipath_etc_conf_cache_FILE:
				multipath_etc_conf_cache_FILE.close()
		else:
#			vi_errorAndLog_obj.vi_print_dbg( 0, 'cached multipath_exe_conf' )
			return mpConf_default_callout, mpConf_blacklistInfo, mpConf_blacklistExceptionsInfo, mpConf_devSpec_LIST

	mpc_contents_LIST, read_failed = st_utils.vi_read_text_lines( vi_errorAndLog_obj, multipathConfFilePath )
	if read_failed:
		vi_errorAndLog_obj.vi_print_err( "ERROR:  %s could not be read !?!" % ( multipathConfFilePath ) )
		return None, None, None, None

	numOfParsingErrs, mpConf_default_callout, mpConf_blacklistInfo, mpConf_blacklistExceptionsInfo, mpConf_devSpec_LIST	= parse_multipath_conf_lines_list( vi_errorAndLog_obj, mpc_contents_LIST )
	if numOfParsingErrs > 0:
		return None, None, None, None
	
	if 0:
		vi_errorAndLog_obj.vi_print_dbg( 0, '' )
		vi_errorAndLog_obj.vi_print_dbg( 0, "Default: %s" % ( mpConf_default_callout ) )
		vi_errorAndLog_obj.vi_print_dbg( 0, '' )
		vi_errorAndLog_obj.vi_print_dbg( 0, "Blacklist:" )
		for blacklist in mpConf_blacklistInfo.mDevNode_LIST:
			vi_errorAndLog_obj.vi_print_dbg( 0, "  d %s" % ( blacklist ) )
		for blacklist in mpConf_blacklistInfo.mDevice_LIST:
			vi_errorAndLog_obj.vi_print_dbg( 0, "  D %s" % ( blacklist ) )
		for blacklist in mpConf_blacklistInfo.mWWID_LIST:
			vi_errorAndLog_obj.vi_print_dbg( 0, "  w %s" % ( blacklist ) )
		vi_errorAndLog_obj.vi_print_dbg( 0, '' )
		vi_errorAndLog_obj.vi_print_dbg( 0, "Blacklist Exceptions:" )
		for blacklist in mpConf_blacklistExceptionsInfo.mDevNode_LIST:
			vi_errorAndLog_obj.vi_print_dbg( 0, "  de %s" % ( blacklist ) )
		for blacklist in mpConf_blacklistExceptionsInfo.mDevice_LIST:
			vi_errorAndLog_obj.vi_print_dbg( 0, "  De %s" % ( blacklist ) )
		for blacklist in mpConf_blacklistExceptionsInfo.mWWID_LIST:
			vi_errorAndLog_obj.vi_print_dbg( 0, "  we %s" % ( blacklist ) )
		vi_errorAndLog_obj.vi_print_dbg( 0, '' )
		vi_errorAndLog_obj.vi_print_dbg( 0, "Devices:" )
		for mpConf_devSpec in mpConf_devSpec_LIST:
			vi_errorAndLog_obj.vi_print_dbg( 0, "    v: '%s' p: '%s'" % ( mpConf_devSpec.mVendor, mpConf_devSpec.mProduct ) )
			vi_errorAndLog_obj.vi_print_dbg( 0, "    [%s]" % ( mpConf_devSpec.mCallout ) )
			if mpConf_devSpec.mBlacklist:
				vi_errorAndLog_obj.vi_print_dbg( 0, "    <%s>" % ( mpConf_devSpec.mBlacklist ) )
			vi_errorAndLog_obj.vi_print_dbg( 0, '' )
		vi_errorAndLog_obj.vi_print_dbg( 0, '' )
		vi_errorAndLog_obj.vi_stop()

	if False and multipathConfFilePath == '/etc/multipath.conf':	# Unnecessary for now
		multipath_etc_conf_cache_FILE		= None
		try:
			multipath_etc_conf_cache_FILE	= open( global_multipath_etc_conf_cache_path, 'wb' )
			pickle.dump( ( mpConf_default_callout, mpConf_blacklistInfo, mpConf_blacklistExceptionsInfo, mpConf_devSpec_LIST ), multipath_etc_conf_cache_FILE, -1 )
		except:
			if multipath_etc_conf_cache_FILE:
				multipath_etc_conf_cache_FILE.close()

	return mpConf_default_callout, mpConf_blacklistInfo, mpConf_blacklistExceptionsInfo, mpConf_devSpec_LIST

####################################################################################################
def get_multipath_conf( vi_errorAndLog_obj ):
	"""
	This routine
	"""

	stdout_LIST, stderr_LIST, exitStatus, deltaTime = st_utils.vi_exe_sync( vi_errorAndLog_obj, "multipathd -k'show config'" )
	del( stderr_LIST )
	mpc_contents_LIST = map( str.strip, stdout_LIST )
	del( stdout_LIST )

	numOfParsingErrs, mpConf_default_callout, mpConf_blacklistInfo, mpConf_blacklistExceptionsInfo, mpConf_devSpec_LIST	= parse_multipath_conf_lines_list( vi_errorAndLog_obj, mpc_contents_LIST )	
	if numOfParsingErrs > 0:
		return None, None, None, None

	etc_default_callout, etc_blacklistInfo, etc_blacklistExceptionsInfo, etc_devSpec_LIST	= parse_multipath_conf_file( vi_errorAndLog_obj, None )
	for devSpec in mpConf_devSpec_LIST:
		if devSpec in etc_devSpec_LIST:
			devSpec.mIsExplicit	= True

	if 0:
		vi_errorAndLog_obj.vi_print_dbg( 0, '' )
		vi_errorAndLog_obj.vi_print_dbg( 0, "Default: %s" % ( mpConf_default_callout ) )
		vi_errorAndLog_obj.vi_print_dbg( 0, '' )
		vi_errorAndLog_obj.vi_print_dbg( 0, "Blacklist:" )
		for blacklist in mpConf_blacklistInfo.mDevNode_LIST:
			vi_errorAndLog_obj.vi_print_dbg( 0, "  d %s" % ( blacklist ) )
		for blacklist in mpConf_blacklistInfo.mDevice_LIST:
			vi_errorAndLog_obj.vi_print_dbg( 0, "  D %s" % ( blacklist ) )
		for blacklist in mpConf_blacklistInfo.mWWID_LIST:
			vi_errorAndLog_obj.vi_print_dbg( 0, "  w %s" % ( blacklist ) )
		vi_errorAndLog_obj.vi_print_dbg( 0, '' )
		vi_errorAndLog_obj.vi_print_dbg( 0, "Blacklist Exceptions:" )
		for blacklist in mpConf_blacklistExceptionsInfo.mDevNode_LIST:
			vi_errorAndLog_obj.vi_print_dbg( 0, "  de %s" % ( blacklist ) )
		for blacklist in mpConf_blacklistExceptionsInfo.mDevice_LIST:
			vi_errorAndLog_obj.vi_print_dbg( 0, "  De %s" % ( blacklist ) )
		for blacklist in mpConf_blacklistExceptionsInfo.mWWID_LIST:
			vi_errorAndLog_obj.vi_print_dbg( 0, "  we %s" % ( blacklist ) )
		vi_errorAndLog_obj.vi_print_dbg( 0, '' )
		vi_errorAndLog_obj.vi_print_dbg( 0, "Devices:" )
		for mpConf_devSpec in mpConf_devSpec_LIST:
			if mpConf_devSpec.mIsExplicit == False:
				vi_errorAndLog_obj.vi_print_dbg( 0, "    -v: '%s' p: '%s'" % ( mpConf_devSpec.mVendor, mpConf_devSpec.mProduct ) )
			else:
				vi_errorAndLog_obj.vi_print_dbg( 0, "    +v: '%s' p: '%s'" % ( mpConf_devSpec.mVendor, mpConf_devSpec.mProduct ) )
			vi_errorAndLog_obj.vi_print_dbg( 0, "    [%s]" % ( mpConf_devSpec.mCallout ) )
			if mpConf_devSpec.mBlacklist:
				vi_errorAndLog_obj.vi_print_dbg( 0, "    <%s>" % ( mpConf_devSpec.mBlacklist ) )
			vi_errorAndLog_obj.vi_print_dbg( 0, '' )
		vi_errorAndLog_obj.vi_print_dbg( 0, '' )
		vi_errorAndLog_obj.vi_stop()

    
	return mpConf_default_callout, mpConf_blacklistInfo, mpConf_blacklistExceptionsInfo, mpConf_devSpec_LIST

####################################################################################################

def ensure_multipath_maps (vi_errorAndLog_obj):
	"""
	Ensures that the sizes of the multipath maps are correct.
	"""

	multipath_maps_DICT = get_multipath_info(vi_errorAndLog_obj, True)

	multipath_map_work_LIST = []

	for multipath_map_name, blkDev_DICT in multipath_maps_DICT.iteritems():
		# Get the device mapper size.
		dm_size = -1
		resize = False
		for blkDev, info in blkDev_DICT.iteritems():
			if dm_size == -1:
				dm_size = get_block_device_size (vi_errorAndLog_obj, "dm-%s" % (info.mDevMapNum))

			if dm_size != info.mSizeInSect:
				resize = True
				vi_errorAndLog_obj.vi_print_dbg(2,"%s: %d != %d (%s)" % (multipath_map_name, dm_size, info.mSizeInSect, blkDev))

		if resize:
			multipath_map_work_LIST.append(multipath_map_name)

	vi_errorAndLog_obj.vi_print_dbg(1,"Work list: %s" % (multipath_map_work_LIST))
	if len(multipath_map_work_LIST) > 0:
		return resize_multipath_maps(vi_errorAndLog_obj, multipath_map_work_LIST)

	return 0

####################################################################################################
def resize_multipath_maps( vi_errorAndLog_obj, multipath_map_LIST ):
	"""
	This routine resizes the given multipath maps.
	"""

	rc = 0

	if len(multipath_map_LIST) <= 0:
		return -1

	for map_name in multipath_map_LIST:
			# Kick multipathd to resize the map.
			exe_string = "multipathd -k'resize map %s'" % (map_name)
			stdout_LIST, stderr_LIST, exitStatus, deltaTime =  \
				st_utils.vi_exe_sync( vi_errorAndLog_obj, exe_string )
			if exitStatus:
				vi_errorAndLog_obj.vi_print_wrn( "%s: Failed to resize map! rc: %d\nstdout:[%s]\nstderr:[%s]" \
							% ( map_name, exitStatus, ''.join( stdout_LIST ), ''.join (stderr_LIST)))
				rc = 1
				continue

			vi_errorAndLog_obj.vi_print_log("%s: Resized." % ( map_name ))

	return rc
####################################################################################################
# END
