#!/usr/bin/env python
#
# Copyright (C) 2011-2014 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.
####################################################################################################
#
#	Author:		Joshua Nicholas
#
"""Utilities for dealing with the open-iscsi program
"""

import st_utils
import st_os

import os
import re
import errno
import traceback
import time

from blockDeviceOperations	import *
from SG_device				import *
from udev				import *
from multipath				import *

####################################################################################################

__all__ = [	'openISCSIstate_OBJ'
		,	'openISCSI_BlkDev_Operations'
		,	'get_openiscsi_session_num_for_blk_dev'
		,	'iscsiadm_exe_sync'
		,	'ensure_openiscsi_session'
		,	'discover_portal'
		,	'login_on_node'
		,	'get_iqn_list'
		,	'get_session_num_to_blk_dev_list_dict'
		,	'get_openiscsi_session_info_dict'
		,	'get_openiscsi_node_list'
		,	'get_openiscsi_iqn_to_blk_dev_list_dict'
		,	'get_openiscsi_iqn_to_sg_dev_list_dict'
		,	'get_blkdev_to_openiscsi_session_info_dict'
		,	'get_sgdev_to_openiscsi_session_info_dict' ]

####################################################################################################

class openISCSIstate_OBJ:

	INACTIVE_SESSION_NUM		= -1

	def __init__( self ):
		self.mSessionNum    = -1
		self.mSCSIHostNum   = -1
		self.mTargetPort    = -1
		self.mTpgt          = -1
		self.mTargetIP      = None
		self.mIQN           = None
		self.mBlkDev_LIST	= []
		self.mSGDev_LIST	= []

	def __repr__( self ):
		return "mSessionNum: %d mSCSIHostNum: %d mTargetPort: %d mTpgt: %d mTargetIP: %s mIQN: %s mBlkDevs: %s mSGDevs: %s" % (
						self.mSessionNum
					,	self.mSCSIHostNum
					,	self.mTargetPort
					,	self.mTpgt
					,	self.mTargetIP
					,	self.mIQN
					,	self.mBlkDev_LIST
					,	self.mSGDev_LIST)

	def __str__ (self) :
		return "%s:%d,%d %s" % (self.mTargetIP, self.mTargetPort, self.mTpgt, self.mIQN)

	def __cmp__( self, other):
		#print "Check (%d:%d, %d:%d, %s:%s)" % (self.mTargetPort, other.mTargetPort, self.mTpgt, other.mTpgt, self.mIQN, other.mIQN)
		if self.mTargetIP < other.mTargetIP:
			return -1
		if self.mTargetPort < other.mTargetPort:
			return -1
		if self.mTpgt < other.mTpgt:
			return -1
 		if self.mIQN < other.mIQN:
			return -1
		if self.mTargetIP > other.mTargetIP:
			return 1
		if self.mTargetPort > other.mTargetPort:
			return 1
		if self.mTpgt > other.mTpgt:
			return 1
		if (self.mIQN > other.mIQN):
			return 1
		return 0

####################################################################################################
class openISCSI_BlkDev_Operations(BlkDev_Operations):

	"""openISCSI class for BlkDev_Operations
	"""

	########################################
	def __init__( self, blkDevInfo, openISCSIstate ):

		if blkDevInfo.mDriver != 'iscsi_tcp':
			raise ValueError, "given blkDevInfo %s is not openISCSI [%s] !?!" % ( blkDevInfo.mBlkDev, blkDevInfo.mDriver )

		self.mBlkDevInfo		= blkDevInfo
		self.mOpenISCSIstate	= openISCSIstate

	########################################
	def __repr__( self ):

		return "openISCSI_BlkDev_Operations[%s]" % ( self.mBlkDevInfo.mBlkDev )

	########################################
	def disconnect_block_device( self, vi_errorAndLog_obj, forceDisconnection ):
		"""
		"""
	#
		std_disconnect_block_device( vi_errorAndLog_obj, self.mBlkDevInfo, forceDisconnection )
		return 0

	########################################
	def revalidate_block_device( self, vi_errorAndLog_obj ):
		"""
		Updates a standard block device to the latests sysfs values
		"""
	#
		retVal	= std_revalidate_block_device( vi_errorAndLog_obj, self.mBlkDevInfo )
		if retVal:
			return retVal

		return 0

####################################################################################################

global__cached__session_num_to_blk_dev_list_DICT			= None
global__cached__openiscsi_session_info_DICT					= None
global__cached__openiscsi_node_info_LIST					= None
global__cached__openiscsi_iqn_DICT							= None
global__cached__openiscsi_iqn_DICT_sg						= None
global__cached__blk_dev_to_openiscsi_session_info_DICT		= None
global__cached__sg_dev_to_openiscsi_session_info_DICT		= None
global__cached__openiscsi_iqn_LIST							= None

####################################################################################################
def get_openiscsi_session_num_for_blk_dev( vi_errorAndLog_obj, blkdev ):
	"""
	This routine gets the openiscsi session num from a blkdev
	"""
#
	blkdev_devLink_path = "/sys/block/%s/device" % ( blkdev )
	if not os.path.exists( blkdev_devLink_path ):
		return -1

	try:
		blkdev_devLink	= os.readlink( blkdev_devLink_path )
	except:
		vi_errorAndLog_obj.vi_print_wrn( "blkdev '%s' vanished [%s] !?!" % ( blkdev, blkdev_devLink_path ) )
		return -1

	try:
		theSCSIid					= re.sub( '^.*/', '', blkdev_devLink )
		host, channel, scsiID, lun	= theSCSIid.split( ':' )
		targetID					= "target%s:%s:%s" % ( host, channel, scsiID  )
		devLink_cmpldRegEx			= re.compile( "[\./]*(.*)/host%s/(.*)%s/%s$" % ( host, targetID, theSCSIid ) )
		devLink_match				= devLink_cmpldRegEx.match( blkdev_devLink )
		specificTransport			= devLink_match.group( 2 ).rstrip( '/' )
		specificTransport_match 	= re.match( '^session(\d+)$', specificTransport )
		iscsi_sessionNum			= int( specificTransport_match.group( 1 ) )
	except:
		return -1

	return iscsi_sessionNum

####################################################################################################
def iscsiadm_exe_sync( vi_errorAndLog_obj, args, inputData_LIST=None ):
	"""
	This routine manages a safe iscsiadm execution
	"""
#
	checkerData	= ( 600, "iscsiadm -m node" )
	checkerFunc	= st_utils.vi_simple_process_checker

	itr	= 0
	while True:

		stdout_LIST, stderr_LIST, exitStatus, deltaTime = \
			st_utils.vi_exe_async_with_checker( vi_errorAndLog_obj, args, checkerFunc, checkerData, inputData_LIST )

		if exitStatus != errno.ETIMEDOUT:
			break

		itr	+= 1
		if itr > 3:
			vi_errorAndLog_obj.vi_print_wrn( "iscsiadm_exe_sync() failed to restart iSCSI daemon" )
			break

		# If there are other iscsiadm commands wedged they will retry their IPC calls
		# with the new iSCSI daemon. This might led back to the wedge state so we kill
		# those before we restart iSCSI daemon.

		pid_LIST = st_utils.vi_pidof( vi_errorAndLog_obj, "iscsiadm" )
		for	pid in pid_LIST:
			st_utils.vi_terminate_process( vi_errorAndLog_obj, pid, True )

		pid_LIST = st_utils.vi_pidof( vi_errorAndLog_obj, "iscsid" )
		for	pid in pid_LIST:
			st_utils.vi_terminate_process( vi_errorAndLog_obj, pid, True )

		vi_errorAndLog_obj.vi_print_wrn( "iscsiadm_exe_sync() restarting iSCSI daemon" )
		os.system( "iscsid &" )	# system() launches the exe in a shell
		time.sleep( 5 )			# Give iscsid a chance to start running

	return ( stdout_LIST, stderr_LIST, exitStatus, deltaTime )

####################################################################################################
def iscsiadm_discovery( vi_errorAndLog_obj, portal_IP_and_port_addr):
	"""
	Invoke the 'iscsiadm' command to execute a iSCSI Discovery Phase
	on the specified IP address, or the IP address and port ("IP:Port").

	Returns:
	Tuple of: return code from iscsiadm, standard output list.

	"""

	exitStatus = errno.ENODEV
	return_LIST = []
	if not portal_IP_and_port_addr:
		return exitStatus,return_LIST

	args_LIST = ['iscsiadm','-m','discovery','-t','st','-p', portal_IP_and_port_addr]
	stdout_LIST, stderr_LIST, exitStatus, deltaTime = \
		iscsiadm_exe_sync( vi_errorAndLog_obj, args_LIST)
	if exitStatus != 0:
		vi_errorAndLog_obj.vi_print_wrn("iscsiadm: Discovery failed for %s\n%s" % \
			(portal_IP_and_port_addr, "".join(stdout_LIST).join(stderr_LIST)))
	else:
		return_LIST = stdout_LIST
	return exitStatus, return_LIST

####################################################################################################
def iscsiadm_rescan( vi_errorAndLog_obj, session):
	"""
	Invoke the 'iscsiadm' command so that it can perform an 'Rescan' of the SCSI devices
	served from the Open-ISCSI host. The iscsiadm command: a) iterates over all the Open-ISCSI
	hosts calling the 'scan' SysFS method with '- - -' causing the kernel to call 'scsi_scan_target'
	which executes SCSI LUNs cmd and compares the list of LUNs with the ones in SysFS
	and adds missing ones doing a full fledge discovery (SCSI INQ, SCSI READ CAPACITT, SCSI
	READ CACHE, etc), however it does _not_ update the LUN information in the SysFS;
 	b) iterate over all of the SCSI devices calling the 'rescan' method which causes the kernel
	to execute an SCSI READ CAPACITY on all of the LUNs and update the sysfs with only this data.


	Returns:
	Tuple of: return code from iscsiadm, standard output list.
	"""

	exitStatus = errno.ENODEV
	return_LIST = []
	if not session:
		return exitStatus, return_LIST

	args_LIST = ['iscsiadm', '-m', 'node', '-R', '-T', session.mIQN]
	stdout_LIST, stderr_LIST, exitStatus, deltaTime = iscsiadm_exe_sync( vi_errorAndLog_obj, args_LIST)
	if len(stderr_LIST) > 0 or exitStatus != 0:
		vi_errorAndLog_obj.vi_print_wrn("Failed to invoke iscsiadm -m node -R --portal %s!" % (session.mIQN))
		return_LIST = stderr_LIST

	return exitStatus,return_LIST

####################################################################################################
def iscsiadm_login(vi_errorAndLog_obj, node):
	"""
	Invoke the 'iscsiadm' command to execute a iSCSI Login Phase
	on to the passed in OpenISCSIstate_OBJ node.

	Returns:
	Tuple of: return code from iscsiadm, standard output list.
	"""

	exitStatus = errno.ENODEV
	return_LIST = []
	if not node:
		return exitStatus,return_LIST

	args_LIST = ['iscsiadm', '-m', 'node', '--targetname', node.mIQN, '--portal', node.mTargetIP, '--login']
	stdout_LIST, stderr_LIST, exitStatus, deltaTime = iscsiadm_exe_sync( vi_errorAndLog_obj, args_LIST)

	if len(stderr_LIST) > 0 or exitStatus != 0:
		vi_errorAndLog_obj.vi_print_wrn( "Login failed : targetname %s portal %s " % (node.mIQN, node.mTargetIP))
		return_LIST = stderr_LIST

	return exitStatus,return_LIST

################################################################################################
def discover_portal( vi_errorAndLog_obj, portal_IP_and_port_str):
	"""
	Executes a iSCSI Discovery Phase task on 'ping' reachable portals. This routine
	takes care of checking whether the the portal is 'ping' reachable. Upon
	completion returns a tuple consisting of the return code along with a list of targets
	discovered. Errors or success states are logged with the vi_errorAndLog_obj object.

	Returns:
	Zero on success.
	"""

	exitStatus = errno.ENODEV
	if not portal_IP_and_port_str:
		vi_errorAndLog_obj.vi_print_wrn("Invalid portal: %s!" % (portal_IP_and_port_str))
		return exitStatus

	portal_IP, portal_Port = vi_os.vi_extract_IP_addr(portal_IP_and_port_str)
	# It is OK to not have an port. Will use the default one.
	if not portal_IP:
		vi_errorAndLog_obj.vi_print_wrn("Malformed IP address of the  portal: %s!" % (portal_IP_and_port_str))
		return exitStatus

	if not vi_os.vi_is_subnet_for_IP_present(vi_errorAndLog_obj, True, portal_IP):
		vi_errorAndLog_obj.vi_print_wrn("ERROR: The subnet route for the iSCSI target %s is NOT available!" % (portal_IP))
		return exitStatus

	vi_errorAndLog_obj.vi_print_log("Checking network access to iSCSI portal :    %s" % (portal_IP_and_port_str))

	# Perform the Open-ISCSI discovery, if applicable
	if vi_os.vi_ping_test(vi_errorAndLog_obj, portal_IP, 60) == 0:
		vi_errorAndLog_obj.vi_print_log("Discover available iSCSI target(s) on portal %s" % (portal_IP_and_port_str))
		exitStatus, targets_LIST = iscsiadm_discovery(vi_errorAndLog_obj, portal_IP_and_port_str)
		if targets_LIST and exitStatus == 0:
			for target in targets_LIST:
				vi_errorAndLog_obj.vi_print_log("%s" % (target.strip()))
	else:
		vi_errorAndLog_obj.vi_print_wrn("iSCSI portal unreachable: %s!" % (portal_IP_and_port_str))

	return exitStatus

####################################################################################################
def ensure_openiscsi_session(vi_errorAndLog_obj, session):
	"""
	Ensures that the SCSI devices provided via the Open-iSCSI module are up-to-date.
	This entails a) scanning each of the SCSI devices of the Open-iSCSI host to check
	if the SCSI INQ commands report the same data as the Sys FS; b) invoking
	the 'iscsiadm -m node -R 'command, which passes '- - -' to the SysFS 'scan' attribute
	thereby invoking the kernel's 'scan_scsi_target(..)' which performs a full
	scan of the HBA detecting new devices and re-generating the SySFS entries.

	Returns:
	Zero on success.
	"""

	exitStatus = errno.ENODEV
	if not session:
		return exitStatus

	# a) Update all of the SG devices based on the Open-iSCSI host number.
	vi_errorAndLog_obj.vi_print_log( "Re-scanning iSCSI target(s) LUN(s) for %s portal" % ( session.mTargetIP ) )
	disconnected_cnt, rescanned_cnt = disconnect_stale_sg_devices_on_host(vi_errorAndLog_obj, session.mSCSIHostNum)
	if disconnected_cnt > 0 or rescanned_cnt > 0:
		if udev_idle( vi_errorAndLog_obj, 3, 120 ):
			vi_errorAndLog_obj.vi_print_wrn( "Timed out on disconnecting SCSI devices on SCSI host: %d!" % (session.mSCSIHostNum))

	# b) invoke 'iscsiadm -m node -R'
	vi_errorAndLog_obj.vi_print_log( "Re-scanning session [sid: %d, target: %s, portal: %s:%d]" % \
		( session.mSessionNum,session.mIQN, session.mTargetIP, session.mTargetPort ) )
	exitStatus, stdout_LIST = iscsiadm_rescan(vi_errorAndLog_obj, session)

	if exitStatus == 0 and stdout_LIST:
		for line in stdout_LIST:
			vi_errorAndLog_obj.vi_print_log("%s" % (line.strip()))

	# c) and wait for a min of 3 seconds for the new disks (if any) to appear.
	if udev_idle( vi_errorAndLog_obj, 3, 120 ):
		vi_errorAndLog_obj.vi_print_wrn( "Timed out on rescanning SCSI host: %d!" % (session.mSCSIHostNum))

	return exitStatus

####################################################################################################
def login_on_node(vi_errorAndLog_obj, record):
	"""
	Executes an iSCSI Login Phase on to the specified openISCSIstate_OBJ node.
	This routine performs: a) a check if a subnet destination route exists for the node;
	b) 'ping' test to the specified portal IP address; c) and, performs the actual
	iSCSI Login Phase using the 'iscsiadm' command.

	Return:
	Zero on success.
	"""

	exitStatus = errno.ENODEV
	if not vi_os.vi_is_subnet_for_IP_present(vi_errorAndLog_obj, True, record.mTargetIP):
		vi_errorAndLog_obj.vi_print_wrn("ERROR: The subnet route for the iSCSI target %s is NOT available!" % (record.mTargetIP))
		return exitStatus

	vi_errorAndLog_obj.vi_print_log( "Checking network access to iSCSI portal : %s " % (record.mTargetIP))
	if vi_os.vi_ping_test(vi_errorAndLog_obj, record.mTargetIP, 60) != 0:
		vi_errorAndLog_obj.vi_print_wrn("ERROR: iSCSI target %s unreachable!" % (record.mTargetIP))
		return exitStatus

	vi_errorAndLog_obj.vi_print_log( "Commence log in to iSCSI target %s @ %s " % (record.mIQN, record.mTargetIP))
	exitStatus, stdout_LIST=iscsiadm_login(vi_errorAndLog_obj, record)
	if exitStatus == 0 and stdout_LIST:
		vi_errorAndLog_obj.vi_print_log( "Login to %s @ %s " % (record.mIQN, record.mTargetIP))
		for line in stdout_LIST:
			vi_errorAndLog_obj.vi_print_log("%s" % (line.strip()))

	return exitStatus

####################################################################################################
def get_iqn_list( vi_errorAndLog_obj, refresh ):
	"""
	This routines returns a list of the currently logged in IQNs.
	"""
	global global__cached__openiscsi_iqn_LIST
	if not refresh and global__cached__openiscsi_iqn_LIST:
		return global__cached__openiscsi_iqn_LIST

	openiscsi_session_info_DICT	= get_openiscsi_session_info_dict( vi_errorAndLog_obj, refresh )

	openiscsi_IQN_LIST = []
	for openiscsi_session_num, openISCSIstate in openiscsi_session_info_DICT.iteritems():
		openiscsi_IQN_LIST		= openISCSIstate.mIQN

	global__cached__openiscsi_iqn_LIST	= openiscsi_IQN_LIST
	return global__cached__openiscsi_iqn_LIST

####################################################################################################
def get_session_num_to_blk_dev_list_dict( vi_errorAndLog_obj, refresh ):
	"""
	This routine returns the dictionary of session numbers to block devices, as so:
		1 -> (sda, sdb, sdau, sdaz)
		2 -> (sdb)
		3 -> ...

	"""
#
	global global__cached__session_num_to_blk_dev_list_DICT
	if not refresh and global__cached__session_num_to_blk_dev_list_DICT:
		return global__cached__session_num_to_blk_dev_list_DICT

	session_num_to_blk_dev_list_DICT	= {}
	for blkdev in os.listdir( '/sys/block' ):
		session_num					= get_openiscsi_session_num_for_blk_dev( vi_errorAndLog_obj, blkdev )
		if session_num < 0:
			continue
		session_num_to_blk_dev_LIST	= session_num_to_blk_dev_list_DICT.get( session_num, None )
		if not session_num_to_blk_dev_LIST:
			session_num_to_blk_dev_list_DICT[ session_num ]	= [ blkdev ]
		else:
			session_num_to_blk_dev_LIST.append( blkdev )

	global__cached__session_num_to_blk_dev_list_DICT	= session_num_to_blk_dev_list_DICT
	return global__cached__session_num_to_blk_dev_list_DICT

####################################################################################################
def get_openiscsi_session_info_dict( vi_errorAndLog_obj, refresh ):
	"""
	This routine extracts the Open-iSCSI session data. The returned values are in the
	dictionary wherein the key is the session number and the value the session data
	(as the openISCSIstate_OBJ)
	"""

	global global__cached__openiscsi_session_info_DICT
	if not refresh and global__cached__openiscsi_session_info_DICT:
		return global__cached__openiscsi_session_info_DICT

	openiscsi_session_info_DICT = {}

	# This is the old code. Not used anymore..
	if 0:
		if not st_utils.vi_pidof( vi_errorAndLog_obj, 'iscsid' ):
			return {}

		openiscsi_session_info_DICT = {}

		#tcp: [1] 10.1.31.241:3260,2 iqn.1992-08.com.netapp:sn.84261533
		#tcp: [2] 10.1.31.240:3260,1 iqn.1992-08.com.netapp:sn.84261533

		openISCSIstate_cmpldRegEx = re.compile( "^tcp: \[(\d+)\]\s+(\S*),(\d+)\s+(\S+)$" )

		session_num_to_blk_dev_list_DICT	= get_session_num_to_blk_dev_list_dict( vi_errorAndLog_obj, refresh )

		try:
			stdout_LIST, stderr_LIST, exitStatus, deltaTime = iscsiadm_exe_sync( vi_errorAndLog_obj, 'iscsiadm -m session' )
			del( stderr_LIST )
			for line in stdout_LIST:
				openISCSIstate_match       	= openISCSIstate_cmpldRegEx.match( line.strip() )
				if not openISCSIstate_match:
					continue
				openISCSIstate             	= openISCSIstate_OBJ()
				openISCSIstate.mSessionNum 	= int(openISCSIstate_match.group(1))
				openISCSIstate.mTargetPort 	= int(openISCSIstate_match.group(2))
				openISCSIstate.mTpgt       	= int(openISCSIstate_match.group(3))
				openISCSIstate.mTargetIP   	= openISCSIstate_match.group(2)
				openISCSIstate.mIQN        	= openISCSIstate_match.group(4)
				openISCSIstate.mBlkDev_LIST	= session_num_to_blk_dev_list_DICT.get( openISCSIstate.mSessionNum, [] )

				openiscsi_session_info_DICT[ openISCSIstate.mSessionNum ] = openISCSIstate
	#			vi_errorAndLog_obj.vi_print_dbg( 0, openISCSIstate )
		except:
	#		vi_errorAndLog_obj.vi_croak_exception( traceback.format_exc() )
			vi_errorAndLog_obj.vi_print_err( "Problems extracting iscsiadm -m session info\n%s" % ( traceback.format_exc() ) )
			return {}

		global__cached__openiscsi_session_info_DICT	= openiscsi_session_info_DICT
		return global__cached__openiscsi_session_info_DICT

	#
	# Directory format of ../iscsi_session is as follow:
	#	session<X>:
	#		-targetname, recovery_tmo, username, etc..
	#		+ device
	#			-uevent
	#			+connection<Y>:0
	#				+ iscsi_connection:connection<Y>:0 (a sym-link)
	#					+ device (symlink to connection<Y>:0
	#					- address, port, persistentaddress, etc
	#			+power
	#			+target<Z>:0:0
	#				+ Z:A:B:C (the SCSI Host:Bus:Channel:LUN ID)
	#					+ block:<sdM> (symlink to /sys/block/sdM) - optional
	#					+ scsi_generic:<sgN> (symlink to /sys/scsi_generic/sgN)
	#					+ and more dirs...
	#					- delete, type, vendor, model, rescan, etc..
	#

	session_root = '/sys/class/iscsi_session/'
	if not os.path.exists( session_root ):
		return {}

	for session in os.listdir( session_root ):
		openISCSIstate             	= openISCSIstate_OBJ()
		openISCSIstate.mSessionNum 	= int(re.sub( '^.*session', '', session ))
		tpgt,read_failed 		= st_utils.vi_read_text_value( vi_errorAndLog_obj, session_root+ session + "/tpgt" )
		if not read_failed:
			openISCSIstate.mTpgt	= int(tpgt)
		openISCSIstate.mIQN, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, session_root + session + "/targetname" )
		device_root = "/sys/class/iscsi_session/%s/device" % (session)
		# ../../../devices/platform/host5/session1
		link = os.readlink(device_root)
		host_match = re.match("^\S*host(\d+)\S*$", link)
		if host_match:
			openISCSIstate.mSCSIHostNum = int(host_match.group(1))
		for root, dirs, files in os.walk( device_root ):
			for dir in dirs:
				# Extract the block device and SCSI Generic name
				if 'target' in dir:
					# Silly sym-links means that we have to re-do a walk on the .../target<Z>:0:0 directory:
					target_root = session_root + "%s/device/%s/" % (session, dir)
					for root, dirs, files in os.walk(target_root):
						# 12:0:0:1, 12:0:0:3, etc..
						for dir in dirs:
							if 'scsi_generic:' in dir:
								sgname = re.sub( 'scsi_generic:', '', dir)
								openISCSIstate.mSGDev_LIST.append(sgname)
							if 'block:' in dir:
								blkName = re.sub( 'block:', '', dir)
								openISCSIstate.mBlkDev_LIST.append(blkName)
				# Extract the address and port
				if 'connection' in dir:
					iscsi_connection_root = session_root + "%s/device/%s/"  % (session, dir)
					for root, dirs, files in os.walk  ( iscsi_connection_root ):
						for dir2 in dirs:
							if 'iscsi_connection' in dir2:
								port, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, \
							 		iscsi_connection_root + dir2 + "/persistent_port")
								if not read_failed:
									openISCSIstate.mTargetPort = int(port)
								else:
									port, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, \
							 				iscsi_connection_root + dir2 + "/" + dir + "/persistent_port")
									if not read_failed:
										openISCSIstate.mTargetPort = int(port)
									else:
										openISCSIstate.mTargetPort = 3260
								openISCSIstate.mTargetIP, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, \
									iscsi_connection_root + dir2 + "/persistent_address")
								if read_failed:
                                                                        openISCSIstate.mTargetIP, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, \
                                                                                iscsi_connection_root + dir2  + "/" + dir + "/persistent_address")
								if read_failed:
									openISCSIstate.mTargetIP, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, \
										iscsi_connection_root + dir2 + "/address")
								if read_failed:
									openISCSIstate.mTargetIP, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, \
										iscsi_connection_root + dir2  + "/" + dir + "/address")


		openiscsi_session_info_DICT[ openISCSIstate.mSessionNum ] = openISCSIstate
		#vi_errorAndLog_obj.vi_print_dbg( 0, openISCSIstate )

	if 0:
		vi_errorAndLog_obj.vi_print_dbg( 0, openiscsi_session_info_DICT )

	global__cached__openiscsi_session_info_DICT	= openiscsi_session_info_DICT
	return global__cached__openiscsi_session_info_DICT

####################################################################################################
def get_openiscsi_node_list( vi_errorAndLog_obj, refresh ):
	"""
	Retrieve the iSCSI node records. The node records contain all the session
	data (which you can get from 'get_openiscsi_session_info_dict') and possibly
	inactive node records for which iSCSI has not logged (or failed) in yet.

	Returns:
	A list of openISCSIstate_OBJ() objects.
	"""

	global global__cached__openiscsi_node_info_LIST
	if not refresh and global__cached__openiscsi_node_info_LIST:
		return global__cached__openiscsi_node_info_LIST

	openiscsi_node_LIST = []

	#10.1.32.230:3260,1 iqn.2001-05.com.equallogic:6-8a0900-ee1a77001-70a000004c3483d8-konrad20gb-2
	#10.1.33.248:3260,2 iqn.1992-01.com.lsi:1535.600a0b800036e48a0000000047b30d8d

	openISCSInode_cmpldRegEx = re.compile( "^(\S+):(\d+),(\d+)\s+(\S*)$" )

	try:
		stdout_LIST, stderr_LIST, exitStatus, deltaTime = iscsiadm_exe_sync( vi_errorAndLog_obj, ['iscsiadm','-m','node'] )
		del( stderr_LIST )
		for line in stdout_LIST:
			openISCSIstate_match       	= openISCSInode_cmpldRegEx.match( line.strip() )
			if not openISCSIstate_match:
				continue
			openISCSIstate             	= openISCSIstate_OBJ()
			openISCSIstate.mSessionNum 	= openISCSIstate.INACTIVE_SESSION_NUM
			openISCSIstate.mTargetPort 	= int(openISCSIstate_match.group(2))
			openISCSIstate.mTpgt       	= int(openISCSIstate_match.group(3))
			openISCSIstate.mTargetIP   	= openISCSIstate_match.group(1)
			openISCSIstate.mIQN        	= openISCSIstate_match.group(4)
			openISCSIstate.mBlkDev_LIST	= None

			openiscsi_node_LIST.append( openISCSIstate )
#			vi_errorAndLog_obj.vi_print_dbg( 0, openISCSIstate )
	except:
#		vi_errorAndLog_obj.vi_croak_exception( traceback.format_exc() )
		vi_errorAndLog_obj.vi_print_err( "Problems extracting iscsiadm -m mode info\n%s" % ( traceback.format_exc() ) )
		return {}
	global__cached__openiscsi_node_info_LIST = openiscsi_node_LIST
	return global__cached__openiscsi_node_info_LIST

####################################################################################################
def get_openiscsi_iqn_to_blk_dev_list_dict( vi_errorAndLog_obj, refresh ):
	"""
	This routine returns the dictionary wherein the key is the IQN and the value are the
	block devices associated with the IQN, as so:

		iqn.1992-01.com.lsi:1535.600a0b800036e48a0000000047b30d8d -> ['sdaa', 'sdab', 'sdac', 'sdae', 'sdaf',..
		iqn.2009.dumpdata.com:disk1.test01 -> ['sde']

	"""
#
	global global__cached__openiscsi_iqn_DICT
	if not refresh and global__cached__openiscsi_iqn_DICT:
		return global__cached__openiscsi_iqn_DICT

	openiscsi_session_info_DICT	= get_openiscsi_session_info_dict( vi_errorAndLog_obj, refresh )

	openiscsi_IQN_DICT	= {}
	blkDev_LIST		= []
	for openiscsi_session_num, openISCSIstate in openiscsi_session_info_DICT.iteritems():
		blkdev_LIST		= openiscsi_IQN_DICT.get( openISCSIstate.mIQN, None )
		if not blkdev_LIST:
			openiscsi_IQN_DICT[ openISCSIstate.mIQN ]	= openISCSIstate.mBlkDev_LIST
		else:
			blkdev_LIST.extend( openISCSIstate.mBlkDev_LIST )

	for openiscsi_IQN, blkdev_LIST in openiscsi_IQN_DICT.iteritems():
		blkdev_LIST.sort()

	if 0:
		vi_errorAndLog_obj.vi_print_dbg( 0, openiscsi_IQN_DICT )

	global__cached__openiscsi_iqn_DICT	= openiscsi_IQN_DICT
	return global__cached__openiscsi_iqn_DICT

####################################################################################################
def get_openiscsi_iqn_to_sg_dev_list_dict( vi_errorAndLog_obj, refresh ):
	"""
	This routine returns the dictionary wherein the key is the IQN and the value are the
	SCSI generic devices associated with the IQN, as so:

		iqn.1992-01.com.lsi:1535.600a0b800036e48a0000000047b30d8d -> ['sg0', 'sg1', 'sg2', 'sg3', 'sg4',..
		iqn.2009.dumpdata.com:disk1.test01 -> ['sg10']

	"""
#
	global global__cached__openiscsi_iqn_DICT_sg
	if not refresh and global__cached__openiscsi_iqn_DICT_sg:
		return global__cached__openiscsi_iqn_DICT_sg

	openiscsi_session_info_DICT	= get_openiscsi_session_info_dict( vi_errorAndLog_obj, refresh )

	openiscsi_IQN_DICT_sg	= {}
	sgDev_LIST		= []
	for openiscsi_session_num, openISCSIstate in openiscsi_session_info_DICT.iteritems():
		sgDev_LIST		= openiscsi_IQN_DICT_sg.get( openISCSIstate.mIQN, None )
		if not sgDev_LIST:
			openiscsi_IQN_DICT_sg[ openISCSIstate.mIQN ]	= openISCSIstate.mSGDev_LIST
		else:
			sgDev_LIST.extend( openISCSIstate.mSGDev_LIST )

	for openiscsi_IQN, sgDev_LIST in openiscsi_IQN_DICT_sg.iteritems():
		sgDev_LIST.sort()

	if 0:
		vi_errorAndLog_obj.vi_print_dbg( 0, openiscsi_IQN_DICT )

	global__cached__openiscsi_iqn_DICT_sg	= openiscsi_IQN_DICT_sg
	return global__cached__openiscsi_iqn_DICT_sg

####################################################################################################
def get_blkdev_to_openiscsi_session_info_dict( vi_errorAndLog_obj, refresh ):
	"""
	This routine returns the dictionary wherein the key is the block device name and the value is
	the the openiscsi session info as so:

		sdaa -> openISCSIstate_OBJ()
		...

	"""
#
	global global__cached__blk_dev_to_openiscsi_session_info_DICT
	if not refresh and global__cached__blk_dev_to_openiscsi_session_info_DICT:
		return global__cached__blk_dev_to_openiscsi_session_info_DICT

	openiscsi_session_info_DICT	= get_openiscsi_session_info_dict( vi_errorAndLog_obj, refresh )

	blk_dev_to_openiscsi_session_info_DICT	= {}
	for openiscsi_session_num, openISCSIstate in openiscsi_session_info_DICT.iteritems():
		for blkDev in openISCSIstate.mBlkDev_LIST:
			blk_dev_to_openiscsi_session_info_DICT[ blkDev ]	= openISCSIstate

	if 0:
		vi_errorAndLog_obj.vi_print_dbg( 0, blk_dev_to_openiscsi_session_info_DICT )

	global__cached__blk_dev_to_openiscsi_session_info_DICT	= blk_dev_to_openiscsi_session_info_DICT
	return global__cached__blk_dev_to_openiscsi_session_info_DICT

####################################################################################################
def get_sgdev_to_openiscsi_session_info_dict( vi_errorAndLog_obj, refresh ):
	"""
	This routine returns the dictionary wherein the key is the SCSI generic device name and the value is
	the openiscsi session info as so:

		sg1 -> openISCSIstate_OBJ()
		...

	"""
#
	global global__cached__sg_dev_to_openiscsi_session_info_DICT
	if not refresh and global__cached__sg_dev_to_openiscsi_session_info_DICT:
		return global__cached__sg_dev_to_openiscsi_session_info_DICT

	openiscsi_session_info_DICT	= get_openiscsi_session_info_dict( vi_errorAndLog_obj, refresh )

	sg_dev_to_openiscsi_session_info_DICT	= {}
	for openiscsi_session_num, openISCSIstate in openiscsi_session_info_DICT.iteritems():
		for sgDev in openISCSIstate.mSGDev_LIST:
			sg_dev_to_openiscsi_session_info_DICT[ sgDev ]	= openISCSIstate

	if 0:
		vi_errorAndLog_obj.vi_print_dbg( 0, sg_dev_to_openiscsi_session_info_DICT )

	global__cached__sg_dev_to_openiscsi_session_info_DICT	= sg_dev_to_openiscsi_session_info_DICT
	return global__cached__sg_dev_to_openiscsi_session_info_DICT

####################################################################################################
# END
