#!/usr/bin/env python
#
# Copyright (C) 2011-2014 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.
####################################################################################################
#
#	Author:		Joshua Nicholas
#
"""Utilities for dealing with files and file related tasks
"""

import os
import fcntl
import stat
import pickle
import st_utils

####################################################################################################

__all__ = [	'vi_fileLock'
		,	'vi_delete_directory_including_file_contents' ]

####################################################################################################

class vi_fileLock:

	"""Class manages a file lock

	NOTE:  All of the methods are no_throw

	Arguments:	theName		name of the lock ( informational ==> NOT the file name )
				thePath		path to file lock file

	Example:

	masterLock = vi_fileLock( 'master', '/var/lock/master_lock_file' )

	masterLock.CreateLockFile()

	masterLock.AcquireLock( True )
	masterLock.ReleaseLock()

	masterLock.RemoveLockFile()
	"""

	########################################
	def __init__(self, theName, thePath):
		self.mLockName = theName
		self.mLockPath = thePath
		self.mLockFD   = None
	########################################
	def CreateLockFile(self):
		try:
			if not os.path.exists( self.mLockPath ):
				fd = os.open( self.mLockPath, os.O_CREAT )
				os.close( fd )
		except OSError:
			return 1 # return bad path ( such as parent dir was deleted just before trying to create )
		return 0

	########################################
	def RemoveLockFile(self):

		self.ReleaseLock()

		try:
			if os.path.exists( self.mLockPath ):
				os.unlink( self.mLockPath )
		except OSError:
			return 1 # return still "in use"

		return 0

	########################################
	def AcquireLock(self, forWrite_BOOL):

		if self.mLockFD:
			return -1 # return already locked as an error

		try:
			if forWrite_BOOL:
				self.mLockFD = os.open( self.mLockPath, os.O_RDWR | os.O_SYNC)
			else:
				self.mLockFD = os.open( self.mLockPath, os.O_RDONLY | os.O_SYNC)
		except OSError:
			return 1 # return bad path ( such as when it is deleted just before trying to acquire )

		if forWrite_BOOL:
			fcntl.flock( self.mLockFD, fcntl.LOCK_EX ) # blocks trying to acquire lock
		else:
			fcntl.flock( self.mLockFD, fcntl.LOCK_SH ) # blocks trying to acquire lock

		return 0

	########################################
	def TryAcquireLock(self, forWrite_BOOL):

		if self.mLockFD:
			return -1

		mLockFD = None
		try:
			if forWrite_BOOL:
				mLockFD = os.open( self.mLockPath, os.O_RDWR | os.O_SYNC)
			else:
				mLockFD = os.open( self.mLockPath, os.O_RDONLY | os.O_SYNC)
		except OSError:
			return 1 # return bad path ( such as when it is deleted just before trying to acquire )

		if forWrite_BOOL:
			try:
				fcntl.flock( mLockFD, fcntl.LOCK_EX | fcntl.LOCK_NB ) # no block.
			except:
				os.close(mLockFD)
				return 1
		else:
			try:
				fcntl.flock( mLockFD, fcntl.LOCK_SH | fcntl.LOCK_NB ) # no block
			except:
				os.close(mLockFD)
				return 1

		self.mLockFD = mLockFD

		return 0

	########################################
	def write(self, data):

		if self.mLockFD:
			return os.write(self.mLockFD, data)

		raise IOError("Not locked for write")

	########################################
	def readline(self, size):

		st_utils.vi_raise_not_implemented()

	########################################
	def read(self, size):

		if self.mLockFD:
			return os.read(self.mLockFD, size)

		raise IOError("Not locked for read")

	########################################
	def LoadPickle(self):

		data = None
		if self.mLockFD:
			try:
				data = pickle.load(self)
			except:
				pass
			try:
				os.lseek(self.mLockFD, 0, 0)
			except:
				pass

		return data


	########################################
	def SavePickle(self, data):

		if self.mLockFD:
			try:
				os.lseek(self.mLockFD, 0, 0)
				os.ftruncate(self.mLockFD, 0)
				pickle.dump(data, self, -1)	# Using binary mode. Do not need readline.
			except:
				pass

	########################################
	def IsLocked(self):
		if self.mLockFD:
			return True
		return False

	########################################
	def ReleaseLock(self):
		if self.mLockFD:
			fcntl.flock( self.mLockFD, fcntl.LOCK_UN )
			os.close( self.mLockFD )
			self.mLockFD = None

	########################################
	def __del__(self):
		self.ReleaseLock()

	########################################
	def __repr__(self):
		return "n: %s p: %s f: %s" % ( self.mLockName, self.mLockPath, self.mLockFD )

####################################################################################################
def vi_delete_directory_including_file_contents( directoryPath ):
	"""
	This routine deletes a directory and its contents IF it only contains regular files
	
	RETURNS:	1	if there is a non-regular file
				2	if not all the files could be deleted
				3	if all the files were deleted, but the directory could not be deleted
	"""
#
	directoryPath_len = len( directoryPath )
	if directoryPath[ directoryPath_len-1 ] == '/':
		directoryPath = directoryPath[ 0 : directoryPath_len-1 ]

	if not os.path.exists( directoryPath ):
		return 0

	filesToDelete_LIST = []
	for file in os.listdir( directoryPath ):
		filePath = directoryPath + '/' + file
		mode = os.stat( filePath )[ stat.ST_MODE ]
		if not stat.S_ISREG(mode):
			return 1
		filesToDelete_LIST.append( filePath )

	isDirNowEmpty = True
	filePathIdx = 0
	filePathIdxEnd_ex = len( filesToDelete_LIST )
	while True:
		try:
			while filePathIdx < filePathIdxEnd_ex:
				filePath = filesToDelete_LIST[ filePathIdx ]
				filePathIdx = filePathIdx + 1
				os.unlink( filePath )
		except:
			isDirNowEmpty = False
			continue
		break

	if not isDirNowEmpty:
		return 2

	try:
		os.rmdir( directoryPath )
	except:
		return 3

	return 0

####################################################################################################
# END
