#!/usr/bin/env python
#
# Copyright (C) 2011-2014 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.
####################################################################################################
#
#	Author:		Joshua Nicholas
#
"""Utilities for parsing the node_configuration_version.xml file.

<NodeConfigurationVersion>

	<ModelVersion>1</ModelVersion>

	<ReleaseVersion>
		<MajorVersion>4</MajorVersion>
		<MinorVersion>5</MinorVersion>
		<PatchVersion>0</PatchVersion>
		<BuildVersion>199</BuildVersion>
	</ReleaseVersion>

	<BuildInfo>20081006_002</BuildInfo>

	<BranchInfo>Sodium:sles64</BranchInfo>

</NodeConfigurationVersion>

"""

import st_utils

import os
import re
import traceback

import xml.dom.minidom

####################################################################################################

__all__ = [	'VI_CONFIGURATION_VERSION__range'
		,	'VI_CONFIGURATION_VERSION__ModelVersion'
		,	'VI_CONFIGURATION_VERSION__ReleaseVersion'
		,	'VI_CONFIGURATION_VERSION__BuildInfo'
		,	'VI_CONFIGURATION_VERSION__BranchInfo'
		,	'vi_configuration_version__base'
		,	'vi_configuration_version__ModelVersion__OBJ'
		,	'vi_configuration_version__ReleaseVersion__OBJ'
		,	'vi_configuration_version__BuildInfo__OBJ'
		,	'vi_configuration_version__BranchInfo__OBJ'
		,	'vi_get_default_configuration_version_fcn_list'
		,	'vi_extract_configuration_version_list' ]

####################################################################################################

VI_CONFIGURATION_VERSION__range				= ( 1, 5 )	# ( inclusive, exclusive ]
VI_CONFIGURATION_VERSION__ModelVersion		= 1
VI_CONFIGURATION_VERSION__ReleaseVersion	= 2
VI_CONFIGURATION_VERSION__BuildInfo			= 3
VI_CONFIGURATION_VERSION__BranchInfo		= 4

class vi_configuration_version__base:

	def GetDirectiveType( self ):
		vi_raise_not_implemented()

	def GetName( self ):
		vi_raise_not_implemented()

	def __init__( self ):
		self.mFcn	= None

####################################################################################################

####################################################################################################
def _INTERNAL__extract_expected_xml_text( node_obj, expectedTagName ):
	"""
	This routine returns the text value for a particular element node
	"""
#
	if node_obj.nodeType != xml.dom.Node.ELEMENT_NODE:
		raise ValueError, "Given node_obj is not an ELEMENT_NODE : %d" % ( node_obj.nodeType )

	if node_obj.tagName != expectedTagName:
		raise ValueError, "Given node_obj.tagName '%s' does NOT match expected '%s'" % ( node_obj.tagName, expectedTagName )

	textNode_LIST	= []
	for child in node_obj.childNodes:
		if child.nodeType == xml.dom.Node.TEXT_NODE:
			textNode_LIST.append( child )
			continue
		if child.nodeType == xml.dom.Node.COMMENT_NODE:
			continue
		raise ValueError, "Given node_obj '%s' contains non-text node : %d : '%s'" % ( node_obj.tagName, child.nodeType, child.toxml() )

	if len( textNode_LIST ) != 1:
		raise ValueError, "Given node_obj '%s' does not have expected single text node : %d" % ( node_obj.tagName, len( textNode_LIST ) )

	return str(textNode_LIST[0].nodeValue)	# nodeValue is Unicode so convert to regular str

####################################################################################################

class vi_configuration_version__ModelVersion__OBJ(vi_configuration_version__base):

	def GetDirectiveType( self ):
		return VI_CONFIGURATION_VERSION__ModelVersion

	def GetName( self ):
		return 'ModelVersion'

	def __init__( self, node_obj ):

		vi_configuration_version__base.__init__( self )

		self.mModelVersion	= _INTERNAL__extract_expected_xml_text( node_obj, 'ModelVersion' ).strip()

	def __repr__( self ):
		return "%s: mModelVersion: %s" % (
				self.GetName()
			,	self.mModelVersion )

####################################################################################################

class vi_configuration_version__ReleaseVersion__OBJ(vi_configuration_version__base):

	def GetDirectiveType( self ):
		return VI_CONFIGURATION_VERSION__ReleaseVersion

	def GetName( self ):
		return 'ReleaseVersion'

	def __init__( self, node_obj ):

		vi_configuration_version__base.__init__( self )

		if node_obj.tagName != 'ReleaseVersion':
			raise ValueError, "Given node_obj.tagName '%s' does NOT match expected 'ReleaseVersion'" % ( node_obj.tagName )

		elemNode_LIST	= []
		for childNode in node_obj.childNodes:
			if childNode.nodeType == xml.dom.Node.ELEMENT_NODE:
				elemNode_LIST.append( childNode )
				continue
			if childNode.nodeType == xml.dom.Node.COMMENT_NODE:
				continue	# Ignore COMMENT_NODEs
			if childNode.nodeType == xml.dom.Node.TEXT_NODE:
				continue	# Ignore TEXT_NODE(whitespace)s
			raise ValueError, "Given ReleaseVersion contains non-text node : %d : '%s'" % ( childNode.nodeType, childNode.toxml() )

		if len( elemNode_LIST ) != 4:
			raise ValueError, "Given ReleaseVersion contains incorrect number of text nodes : %d" % ( len( elemNode_LIST ) )

		self.mMajorVersion	= _INTERNAL__extract_expected_xml_text( elemNode_LIST[0], 'MajorVersion' ).strip()
		self.mMinorVersion	= _INTERNAL__extract_expected_xml_text( elemNode_LIST[1], 'MinorVersion' ).strip()
		self.mPatchVersion	= _INTERNAL__extract_expected_xml_text( elemNode_LIST[2], 'PatchVersion' ).strip()
		self.mBuildVersion	= _INTERNAL__extract_expected_xml_text( elemNode_LIST[3], 'BuildVersion' ).strip()

	def __repr__( self ):
		return "%s: mMajorVersion: %s mMinorVersion: %s mPatchVersion: %s mBuildVersion: %s" % (
				self.GetName()
			,	self.mMajorVersion
			,	self.mMinorVersion
			,	self.mPatchVersion
			,	self.mBuildVersion )

####################################################################################################

class vi_configuration_version__BuildInfo__OBJ(vi_configuration_version__base):

	def GetDirectiveType( self ):
		return VI_CONFIGURATION_VERSION__BuildInfo

	def GetName( self ):
		return 'BuildInfo'

	def __init__( self, node_obj ):

		vi_configuration_version__base.__init__( self )

		self.mBuildInfo	= _INTERNAL__extract_expected_xml_text( node_obj, 'BuildInfo' ).strip()

	def __repr__( self ):
		return "%s: mBuildInfo: %s" % (
				self.GetName()
			,	self.mBuildInfo )

####################################################################################################

class vi_configuration_version__BranchInfo__OBJ(vi_configuration_version__base):

	def GetDirectiveType( self ):
		return VI_CONFIGURATION_VERSION__BranchInfo

	def GetName( self ):
		return 'BranchInfo'

	def __init__( self, node_obj ):

		vi_configuration_version__base.__init__( self )

		self.mBranchInfo	= _INTERNAL__extract_expected_xml_text( node_obj, 'BranchInfo' ).strip()

	def __repr__( self ):
		return "%s: mBranchInfo: %s" % (
				self.GetName()
			,	self.mBranchInfo )

####################################################################################################
def vi_get_default_configuration_version_fcn_list():
	"""
	This routine
	"""
#
	def ret_None(*x):	return None

	configurationVersionFcn_LIST	= map( ret_None, range( VI_CONFIGURATION_VERSION__range[1] ) )

	return configurationVersionFcn_LIST

####################################################################################################
def vi_extract_configuration_version_list( vi_errorAndLog_obj, configuration_version_xml_lines_buffer, configurationVersionFcn_LIST ):
	"""
	This routine parses the given configuration_version_xml_lines_buffer into the list of
	directives (vi_configuration_version__base subclasses) and returns them.
	"""
#
	vi_errorAndLog_obj.vi_print_dbg( 3, "XMLlines: '%s'" % ( configuration_version_xml_lines_buffer ) )

	if len( configurationVersionFcn_LIST ) < VI_CONFIGURATION_VERSION__range[1]:
		vi_errorAndLog_obj.vi_print_err( "Given configurationVersionFcn_LIST is too short : %d" % ( len( configurationVersionFcn_LIST ) ) )
		return []

	try:
		configuration_version_dom	= xml.dom.minidom.parseString( configuration_version_xml_lines_buffer )
	except:
		vi_errorAndLog_obj.vi_print_err( "Failed to parse limits_and_memory_cfgs_xml '%s' : \n%s" % ( configuration_version_xml_lines_buffer, traceback.format_exc() ) )
		return []

	if not configuration_version_dom.firstChild:
		vi_errorAndLog_obj.vi_print_wrn( "vi_extract_configuration_version_list() Not even <NodeConfigurationVersion></NodeConfigurationVersion> specified" )
		return []

	def ret_None(*x):	return None

	configuration_version_component_LIST	= map( ret_None, range( VI_CONFIGURATION_VERSION__range[1] ) )

	for childNode in configuration_version_dom.firstChild.childNodes:

		if childNode.nodeType != xml.dom.Node.ELEMENT_NODE:
			continue	# Ignore COMMENT_NODEs and TEXT_NODE(whitespace)s

		vi_errorAndLog_obj.vi_print_dbg( 3, "limAmemCfg:", childNode.toxml() )

		newTask	= None
		try:
			if   childNode.tagName == 'ModelVersion':
				newTask		= vi_configuration_version__ModelVersion__OBJ( childNode )
			elif childNode.tagName == 'ReleaseVersion':
				newTask		= vi_configuration_version__ReleaseVersion__OBJ( childNode )
			elif childNode.tagName == 'BuildInfo':
				newTask		= vi_configuration_version__BuildInfo__OBJ( childNode )
			elif childNode.tagName == 'BranchInfo':
				newTask		= vi_configuration_version__BranchInfo__OBJ( childNode )
			else:
				vi_errorAndLog_obj.vi_print_wrn( "Unknown NodeConfigurationVersion element node.tagName '%s'" % ( childNode.tagName ) )
		except:
			vi_errorAndLog_obj.vi_print_err( "Failed to create directive for '%s' : '%s' : \n%s" % ( childNode.tagName, childNode.toxml(), traceback.format_exc() ) )

		if newTask:
			newTask.mFcn														= configurationVersionFcn_LIST[ newTask.GetDirectiveType() ]
			configuration_version_component_LIST[ newTask.GetDirectiveType() ]	= newTask

	configuration_version_dom.unlink()	# This MUST be done or the memory will NOT be reclaimed !!!

	return configuration_version_component_LIST

####################################################################################################
# END
