#!/usr/bin/env python
#
# Copyright (C) 2011-2014 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.
####################################################################################################
#
#	Author:		Joshua Nicholas
#
"""Utilities for parsing the limits_and_memory_config.xml file.

<LimitsAndMemoryCfg>

	<!-- ================================================================== -->
	<!-- Storage Connection Limits                                          -->

<!--	<StorageConnectionLimits>											-->
<!--		<MaxDiskPaths>1024</MaxDiskPaths>								-->
<!--		<MaxDisks>512</MaxDisks>										-->
<!--	</StorageConnectionLimits>											-->

	<!-- ================================================================== -->
	<!-- LVM Limits                                                         -->

<!--	<LVMlimits>																-->
<!--		<MaxActiveVolumeGroups>200</MaxActiveVolumeGroups>					-->
<!--		<MaxActiveLogicalVolumes>200</MaxActiveLogicalVolumes>				-->
<!--		<MaxActiveLogicalVolumesPerVG>100</MaxActiveLogicalVolumesPerVG>	-->
<!--		<MaxActiveSnapshots>2</MaxActiveSnapshots>							-->
<!--	</LVMlimits>															-->

	<!-- ================================================================== -->
	<!-- Virtual Machine Limits                                             -->

<!--	<VirtualMachineLimits>												-->
<!--		<MaxVirtualMachines>32</MaxVirtualMachines>						-->
<!--		<MaxVNCconnections>32</MaxVNCconnections>						-->
<!--	</VirtualMachineLimits>												-->

	<!-- ================================================================== -->
	<!-- Dom0 Memory Configuration                                          -->

<!--	<Dom0MemoryCfg>															-->
<!--		<InitialDom0MemoryReserveInMB>100</InitialDom0MemoryReserveInMB>	-->
<!--	</Dom0MemoryCfg>														-->

</LimitsAndMemoryCfg>

"""

import st_utils

import os
import re
import traceback

import xml.dom.minidom

####################################################################################################

__all__ = [	'VI_LIMITS_AND_MEMORY_CFG__range'
		,	'VI_LIMITS_AND_MEMORY_CFG__StorageConnectionLimits'
		,	'VI_LIMITS_AND_MEMORY_CFG__LVMlimits'
		,	'VI_LIMITS_AND_MEMORY_CFG__VirtualMachineLimits'
		,	'VI_LIMITS_AND_MEMORY_CFG__Dom0MemoryCfg'
		,	'vi_limits_and_memory_cfg__base'
		,	'vi_limits_and_memory_cfg__StorageConnectionLimits__OBJ'
		,	'vi_limits_and_memory_cfg__LVMlimits__OBJ'
		,	'vi_limits_and_memory_cfg__VirtualMachineLimits__OBJ'
		,	'vi_limits_and_memory_cfg__Dom0MemoryCfg__OBJ'
		,	'vi_get_limits_and_memory_cfg_each_list'
		,	'vi_get_default_limits_and_memory_cfg_fcn_list'
		,	'vi_extract_limits_and_memory_cfg_list' ]

####################################################################################################

VI_LIMITS_AND_MEMORY_CFG__range						= ( 1, 5 )	# ( inclusive, exclusive ]
VI_LIMITS_AND_MEMORY_CFG__StorageConnectionLimits	= 1
VI_LIMITS_AND_MEMORY_CFG__LVMlimits					= 2
VI_LIMITS_AND_MEMORY_CFG__VirtualMachineLimits		= 3
VI_LIMITS_AND_MEMORY_CFG__Dom0MemoryCfg				= 4

class vi_limits_and_memory_cfg__base:

	def GetDirectiveType( self ):
		vi_raise_not_implemented()

	def GetName( self ):
		vi_raise_not_implemented()

	def __init__( self ):
		self.mFcn	= None

####################################################################################################

####################################################################################################
def _INTERNAL__extract_expected_xml_text( node_obj, expectedTagName ):
	"""
	This routine returns the text value for a particular element node
	"""
#
	if node_obj.nodeType != xml.dom.Node.ELEMENT_NODE:
		raise ValueError, "Given node_obj is not an ELEMENT_NODE : %d" % ( node_obj.nodeType )

	if node_obj.tagName != expectedTagName:
		raise ValueError, "Given node_obj.tagName '%s' does NOT match expected '%s'" % ( node_obj.tagName, expectedTagName )

	textNode_LIST	= []
	for child in node_obj.childNodes:
		if child.nodeType == xml.dom.Node.TEXT_NODE:
			textNode_LIST.append( child )
			continue
		if child.nodeType == xml.dom.Node.COMMENT_NODE:
			continue
		raise ValueError, "Given node_obj '%s' contains non-text node : %d : '%s'" % ( node_obj.tagName, child.nodeType, child.toxml() )

	if len( textNode_LIST ) != 1:
		raise ValueError, "Given node_obj '%s' does not have expected single text node : %d" % ( node_obj.tagName, len( textNode_LIST ) )

	return str(textNode_LIST[0].nodeValue)	# nodeValue is Unicode so convert to regular str

####################################################################################################

class vi_limits_and_memory_cfg__StorageConnectionLimits__OBJ(vi_limits_and_memory_cfg__base):

	def GetDirectiveType( self ):
		return VI_LIMITS_AND_MEMORY_CFG__StorageConnectionLimits

	def GetName( self ):
		return 'StorageConnectionLimits'

	def __init__( self, node_obj ):

		vi_limits_and_memory_cfg__base.__init__( self )

		if node_obj == None:
			self.mMaxDiskPaths	= None
			self.mMaxDisks		= None
			return

		if node_obj.tagName != 'StorageConnectionLimits':
			raise ValueError, "Given node_obj.tagName '%s' does NOT match expected 'StorageConnectionLimits'" % ( node_obj.tagName )

		elemNode_LIST	= []
		for childNode in node_obj.childNodes:
			if childNode.nodeType == xml.dom.Node.ELEMENT_NODE:
				elemNode_LIST.append( childNode )
				continue
			if childNode.nodeType == xml.dom.Node.COMMENT_NODE:
				continue	# Ignore COMMENT_NODEs
			if childNode.nodeType == xml.dom.Node.TEXT_NODE:
				continue	# Ignore TEXT_NODE(whitespace)s
			raise ValueError, "Given StorageConnectionLimits contains non-text node : %d : '%s'" % ( childNode.nodeType, childNode.toxml() )

		if len( elemNode_LIST ) != 2:
			raise ValueError, "Given StorageConnectionLimits contains incorrect number of text nodes : %d" % ( len( elemNode_LIST ) )

		self.mMaxDiskPaths	= _INTERNAL__extract_expected_xml_text( elemNode_LIST[0], 'MaxDiskPaths' ).strip()
		self.mMaxDisks		= _INTERNAL__extract_expected_xml_text( elemNode_LIST[1], 'MaxDisks' ).strip()

	def set_values( self, theMaxDiskPaths, theMaxDisks ):
		self.mMaxDiskPaths	= theMaxDiskPaths
		self.mMaxDisks		= theMaxDisks

	def __repr__( self ):
		return "%s: mMaxDiskPaths: %s mMaxDisks: %s" % (
				self.GetName()
			,	self.mMaxDiskPaths
			,	self.mMaxDisks )

####################################################################################################

class vi_limits_and_memory_cfg__LVMlimits__OBJ(vi_limits_and_memory_cfg__base):

	def GetDirectiveType( self ):
		return VI_LIMITS_AND_MEMORY_CFG__LVMlimits

	def GetName( self ):
		return 'LVMlimits'

	def __init__( self, node_obj ):

		vi_limits_and_memory_cfg__base.__init__( self )

		if node_obj == None:
			self.mMaxActiveVolumeGroups			= None
			self.mMaxActiveLogicalVolumes		= None
			self.mMaxActiveLogicalVolumesPerVG	= None
			self.mMaxActiveSnapshots			= None
			return

		if node_obj.tagName != 'LVMlimits':
			raise ValueError, "Given node_obj.tagName '%s' does NOT match expected 'LVMlimits'" % ( node_obj.tagName )

		elemNode_LIST	= []
		for childNode in node_obj.childNodes:
			if childNode.nodeType == xml.dom.Node.ELEMENT_NODE:
				elemNode_LIST.append( childNode )
				continue
			if childNode.nodeType == xml.dom.Node.COMMENT_NODE:
				continue	# Ignore COMMENT_NODEs
			if childNode.nodeType == xml.dom.Node.TEXT_NODE:
				continue	# Ignore TEXT_NODE(whitespace)s
			raise ValueError, "Given LVMlimits contains non-text node : %d : '%s'" % ( childNode.nodeType, childNode.toxml() )

		if len( elemNode_LIST ) != 4:
			raise ValueError, "Given LVMlimits contains incorrect number of text nodes : %d" % ( len( elemNode_LIST ) )

		self.mMaxActiveVolumeGroups			= _INTERNAL__extract_expected_xml_text( elemNode_LIST[0], 'MaxActiveVolumeGroups' ).strip()
		self.mMaxActiveLogicalVolumes		= _INTERNAL__extract_expected_xml_text( elemNode_LIST[1], 'MaxActiveLogicalVolumes' ).strip()
		self.mMaxActiveLogicalVolumesPerVG	= _INTERNAL__extract_expected_xml_text( elemNode_LIST[2], 'MaxActiveLogicalVolumesPerVG' ).strip()
		self.mMaxActiveSnapshots			= _INTERNAL__extract_expected_xml_text( elemNode_LIST[3], 'MaxActiveSnapshots' ).strip()

	def set_values( self, theMaxActiveVolumeGroups, theMaxActiveLogicalVolumes, theMaxActiveLogicalVolumesPerVG, theMaxActiveSnapshots ):
		self.mMaxActiveVolumeGroups			= theMaxActiveVolumeGroups
		self.mMaxActiveLogicalVolumes		= theMaxActiveLogicalVolumes
		self.mMaxActiveLogicalVolumesPerVG	= theMaxActiveLogicalVolumesPerVG
		self.mMaxActiveSnapshots			= theMaxActiveSnapshots

	def __repr__( self ):
		return "%s: mMaxActiveVolumeGroups: %s mMaxActiveLogicalVolumes: %s mMaxActiveLogicalVolumesPerVG: %s mMaxActiveSnapshots: %s" % (
				self.GetName()
			,	self.mMaxActiveVolumeGroups
			,	self.mMaxActiveLogicalVolumes
			,	self.mMaxActiveLogicalVolumesPerVG
			,	self.mMaxActiveSnapshots )

####################################################################################################

class vi_limits_and_memory_cfg__VirtualMachineLimits__OBJ(vi_limits_and_memory_cfg__base):

	def GetDirectiveType( self ):
		return VI_LIMITS_AND_MEMORY_CFG__VirtualMachineLimits

	def GetName( self ):
		return 'VirtualMachineLimits'

	def __init__( self, node_obj ):

		vi_limits_and_memory_cfg__base.__init__( self )

		if node_obj == None:
			self.mMaxVirtualMachines	= None
			self.mMaxVNCconnections		= None
			return

		if node_obj.tagName != 'VirtualMachineLimits':
			raise ValueError, "Given node_obj.tagName '%s' does NOT match expected 'VirtualMachineLimits'" % ( node_obj.tagName )

		elemNode_LIST	= []
		for childNode in node_obj.childNodes:
			if childNode.nodeType == xml.dom.Node.ELEMENT_NODE:
				elemNode_LIST.append( childNode )
				continue
			if childNode.nodeType == xml.dom.Node.COMMENT_NODE:
				continue	# Ignore COMMENT_NODEs
			if childNode.nodeType == xml.dom.Node.TEXT_NODE:
				continue	# Ignore TEXT_NODE(whitespace)s
			raise ValueError, "Given VirtualMachineLimits contains non-text node : %d : '%s'" % ( childNode.nodeType, childNode.toxml() )

		if len( elemNode_LIST ) != 2:
			raise ValueError, "Given VirtualMachineLimits contains incorrect number of text nodes : %d" % ( len( elemNode_LIST ) )

		self.mMaxVirtualMachines	= _INTERNAL__extract_expected_xml_text( elemNode_LIST[0], 'MaxVirtualMachines' ).strip()
		self.mMaxVNCconnections		= _INTERNAL__extract_expected_xml_text( elemNode_LIST[1], 'MaxVNCconnections' ).strip()

	def set_values( self, theMaxVirtualMachines, theMaxVNCconnections ):
		self.mMaxVirtualMachines	= theMaxVirtualMachines
		self.mMaxVNCconnections		= theMaxVNCconnections

	def __repr__( self ):
		return "%s: mMaxVirtualMachines: %s mMaxVNCconnections: %s" % (
				self.GetName()
			,	self.mMaxVirtualMachines
			,	self.mMaxVNCconnections )

####################################################################################################

class vi_limits_and_memory_cfg__Dom0MemoryCfg__OBJ(vi_limits_and_memory_cfg__base):

	def GetDirectiveType( self ):
		return VI_LIMITS_AND_MEMORY_CFG__Dom0MemoryCfg

	def GetName( self ):
		return 'Dom0MemoryCfg'

	def __init__( self, node_obj ):

		vi_limits_and_memory_cfg__base.__init__( self )

		if node_obj == None:
			self.mInitialDom0MemoryReserveInMB	= None
			return

		if node_obj.tagName != 'Dom0MemoryCfg':
			raise ValueError, "Given node_obj.tagName '%s' does NOT match expected 'Dom0MemoryCfg'" % ( node_obj.tagName )

		elemNode_LIST	= []
		for childNode in node_obj.childNodes:
			if childNode.nodeType == xml.dom.Node.ELEMENT_NODE:
				elemNode_LIST.append( childNode )
				continue
			if childNode.nodeType == xml.dom.Node.COMMENT_NODE:
				continue	# Ignore COMMENT_NODEs
			if childNode.nodeType == xml.dom.Node.TEXT_NODE:
				continue	# Ignore TEXT_NODE(whitespace)s
			raise ValueError, "Given Dom0MemoryCfg contains non-text node : %d : '%s'" % ( childNode.nodeType, childNode.toxml() )

		if len( elemNode_LIST ) != 1:
			raise ValueError, "Given Dom0MemoryCfg contains incorrect number of text nodes : %d" % ( len( elemNode_LIST ) )

		self.mInitialDom0MemoryReserveInMB	= _INTERNAL__extract_expected_xml_text( elemNode_LIST[0], 'InitialDom0MemoryReserveInMB' ).strip()

	def set_values( self, theInitialDom0MemoryReserveInMB ):
		self.mInitialDom0MemoryReserveInMB	= theInitialDom0MemoryReserveInMB

	def __repr__( self ):
		return "%s: mInitialDom0MemoryReserveInMB: %s" % (
				self.GetName()
			,	self.mInitialDom0MemoryReserveInMB )

####################################################################################################
def vi_get_limits_and_memory_cfg_each_list():
	"""
	This routine
	"""
#
	def ret_None(*x):	return None

	limitsAndMemoryCfgEach_LIST	= map( ret_None, range( VI_LIMITS_AND_MEMORY_CFG__range[1] ) )

	limitsAndMemoryCfgEach_LIST[ VI_LIMITS_AND_MEMORY_CFG__StorageConnectionLimits ]	= \
								 vi_limits_and_memory_cfg__StorageConnectionLimits__OBJ( None )

	limitsAndMemoryCfgEach_LIST[ VI_LIMITS_AND_MEMORY_CFG__LVMlimits ]	= \
								 vi_limits_and_memory_cfg__LVMlimits__OBJ( None )

	limitsAndMemoryCfgEach_LIST[ VI_LIMITS_AND_MEMORY_CFG__VirtualMachineLimits ]	= \
								 vi_limits_and_memory_cfg__VirtualMachineLimits__OBJ( None )

	limitsAndMemoryCfgEach_LIST[ VI_LIMITS_AND_MEMORY_CFG__Dom0MemoryCfg ]	= \
								 vi_limits_and_memory_cfg__Dom0MemoryCfg__OBJ( None )

	return limitsAndMemoryCfgEach_LIST

####################################################################################################
def vi_get_default_limits_and_memory_cfg_fcn_list():
	"""
	This routine
	"""
#
	def ret_None(*x):	return None

	limitsAndMemoryCfgFcn_LIST	= map( ret_None, range( VI_LIMITS_AND_MEMORY_CFG__range[1] ) )

	return limitsAndMemoryCfgFcn_LIST

####################################################################################################
def vi_extract_limits_and_memory_cfg_list( vi_errorAndLog_obj, limits_and_memory_cfgs_xml_lines_buffer, limitsAndMemoryCfgFcn_LIST ):
	"""
	This routine parses the given limits_and_memory_cfgs_xml_lines_LIST into the list of
	directives (vi_limits_and_memory_cfg__base subclasses) and returns them.
	"""
#
	vi_errorAndLog_obj.vi_print_dbg( 3, "XMLlines: '%s'" % ( limits_and_memory_cfgs_xml_lines_buffer ) )

	if len( limitsAndMemoryCfgFcn_LIST ) < VI_LIMITS_AND_MEMORY_CFG__range[1]:
		vi_errorAndLog_obj.vi_print_err( "Given limitsAndMemoryCfgFcn_LIST is too short : %d" % ( len( limitsAndMemoryCfgFcn_LIST ) ) )
		return []

	try:
		limits_and_memory_cfgs_dom	= xml.dom.minidom.parseString( limits_and_memory_cfgs_xml_lines_buffer )
	except:
		vi_errorAndLog_obj.vi_print_err( "Failed to parse limits_and_memory_cfgs_xml '%s' : \n%s" % ( limits_and_memory_cfgs_xml_lines_buffer, traceback.format_exc() ) )
		return []

	if not limits_and_memory_cfgs_dom.firstChild:
		vi_errorAndLog_obj.vi_print_wrn( "vi_extract_limits_and_memory_cfg_list() Not even <LimitsAndMemoryCfg></LimitsAndMemoryCfg> specified" )
		return []

	limits_and_memory_cfgs_directive_LIST	= []

	for childNode in limits_and_memory_cfgs_dom.firstChild.childNodes:

		if childNode.nodeType != xml.dom.Node.ELEMENT_NODE:
			continue	# Ignore COMMENT_NODEs and TEXT_NODE(whitespace)s

		vi_errorAndLog_obj.vi_print_dbg( 3, "limAmemCfg:", childNode.toxml() )

		newTask	= None
		try:
			if   childNode.tagName == 'StorageConnectionLimits':
				newTask		= vi_limits_and_memory_cfg__StorageConnectionLimits__OBJ( childNode )
			elif childNode.tagName == 'LVMlimits':
				newTask		= vi_limits_and_memory_cfg__LVMlimits__OBJ( childNode )
			elif childNode.tagName == 'VirtualMachineLimits':
				newTask		= vi_limits_and_memory_cfg__VirtualMachineLimits__OBJ( childNode )
			elif childNode.tagName == 'Dom0MemoryCfg':
				newTask		= vi_limits_and_memory_cfg__Dom0MemoryCfg__OBJ( childNode )
			else:
				vi_errorAndLog_obj.vi_print_wrn( "Unknown LimitsAndMemoryCfg element node.tagName '%s'" % ( childNode.tagName ) )
		except:
			vi_errorAndLog_obj.vi_print_err( "Failed to create directive for '%s' : '%s' : \n%s" % ( childNode.tagName, childNode.toxml(), traceback.format_exc() ) )

		if newTask:
			newTask.mFcn	= limitsAndMemoryCfgFcn_LIST[ newTask.GetDirectiveType() ]
			limits_and_memory_cfgs_directive_LIST.append( newTask )

	limits_and_memory_cfgs_dom.unlink()	# This MUST be done or the memory will NOT be reclaimed !!!

	return limits_and_memory_cfgs_directive_LIST

####################################################################################################
# END
